/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#include "analyze.h"

/**
 * @description: 分析器，进行语义分析和查询重写，需要检查不符合语义规定的部分
 * @param {shared_ptr<ast::TreeNode>} parse parser生成的结果集
 * @return {shared_ptr<Query>} Query 
 */
std::shared_ptr<Query> Analyze::do_analyze(std::shared_ptr<ast::TreeNode> parse,std::vector<std::vector<Value>> final_v)
{
    std::shared_ptr<Query> query = std::make_shared<Query>();
    if (auto x = std::dynamic_pointer_cast<ast::SelectStmt>(parse))
    {
        // 处理表名
        query->tables = std::move(*x->tabs);
        /** TODO: 检查表是否存在 */
        for(const auto&tab:query->tables){
            if(!sm_manager_->fhs_.count(tab)){
                throw TableNotFoundError(tab);
                return nullptr;
            }
        }
        
        // 处理target list，再target list中添加上表名，例如 a.id
        if(x->cols)
        for (auto &sv_sel_col : *x->cols) {
            TabCol sel_col = {
                .tab_name = sv_sel_col->tab_name, 
                .col_name = sv_sel_col->col_name,
                .new_col_name=sv_sel_col->new_name,
                .aggr_type=sv_sel_col->aggr_type};
            query->cols.push_back(sel_col);
        }
        
        std::vector<ColMeta> all_cols;
        get_all_cols(query->tables, all_cols);
        if(x->group){
            for(auto&col:*x->group->cols){
                TabCol group_col = {
                    .tab_name = col->tab_name, 
                    .col_name = col->col_name,
                    .new_col_name=col->new_name,
                    .aggr_type=col->aggr_type};
                query->group_cols.push_back(check_column(all_cols,group_col));
            }
        }
        

        int group_by_status=0;//1=some column is not in group by;2=some column is in group by
        if (query->cols.empty()) {
            // select all columns
            for (auto &col : all_cols) {
                TabCol sel_col = {.tab_name = col.tab_name, .col_name = col.name,.new_col_name="",.aggr_type=ast::AggrType::NO_AGGR};
                query->cols.push_back(sel_col);
                bool in_group=false;
                for(auto&group_col:query->group_cols){
                    if(group_col.col_name==sel_col.col_name&&group_col.tab_name==sel_col.tab_name){
                        sel_col.aggr_type=group_col.aggr_type;
                        in_group=true;
                    }
                }
                if(in_group){
                    if(group_by_status==1){
                        throw AmbiguousColumnError("group by error");
                    }
                    group_by_status=2;
                }
                else{
                    if(group_by_status==2){
                        throw AmbiguousColumnError("group by error");
                    }
                    group_by_status=1;
                }
            }
        } else {
            // infer table name from column name
            for (auto &sel_col : query->cols) {
                if(sel_col.aggr_type==ast::AggrType::COUNT&&sel_col.col_name=="*"){
                    sel_col.col_name=all_cols[0].name;
                    sel_col.tab_name=all_cols[0].tab_name;
                }
                sel_col = check_column(all_cols, sel_col);  // 列元数据校验
                if(sel_col.aggr_type!=ast::AggrType::NO_AGGR){
                    query->use_group=true;
                }
                bool in_group=false;
                if(sel_col.aggr_type!=ast::AggrType::NO_AGGR)in_group=true;
                else
                    for(auto&group_col:query->group_cols){
                        if(group_col.col_name==sel_col.col_name&&group_col.tab_name==sel_col.tab_name){
                            in_group=true;
                        }
                    }
                if(in_group){
                    if(group_by_status==1){
                        throw AmbiguousColumnError("group by error");
                    }
                    group_by_status=2;
                }
                else{
                    if(group_by_status==2){
                        throw AmbiguousColumnError("group by error");
                    }
                    group_by_status=1;
                }
            }
        }
        //处理where条件
        if(!x->conds){
            auto tmp_v=std::vector<std::shared_ptr<ast::BinaryExpr>>{};
            x->conds=std::make_shared<std::vector<std::shared_ptr<ast::BinaryExpr>>>(tmp_v);
        }
        get_clause(*x->conds, query->conds,false);
        int vec_cnt=0;
        for(int i=0;i<query->conds.size();i++){
            if(query->conds[i].is_rhs_vector){
                auto res=final_v[vec_cnt++];
                if(res.size()>1&&query->conds[i].op!=CompOp::OP_IN){
                    throw SonSelStatementError();
                }
                if(res.size()==0){
                    query->conds[i].op=CompOp::OP_IN;
                }
                if(query->conds[i].op!=CompOp::OP_IN){
                    
                    query->conds[i].rhs_val=res[0];
                    query->conds[i].is_rhs_val=true;
                    query->conds[i].is_rhs_vector=false;
                    if(query->conds[i].lhs_col.col_name!="*")query->conds[i].lhs_col = check_column(all_cols, query->conds[i].lhs_col);
                    else continue;
                    TabMeta &lhs_tab = sm_manager_->db_.get_table(query->conds[i].lhs_col.tab_name);
                    auto lhs_col = lhs_tab.get_col(query->conds[i].lhs_col.col_name);
                    ColType lhs_type = lhs_col->type;

                    if(lhs_type==ColType::TYPE_FLOAT&&query->conds[i].rhs_val.type==ColType::TYPE_INT){
                        query->conds[i].rhs_val.float_val=(float)query->conds[i].rhs_val.int_val;
                        *(float*)query->conds[i].rhs_val.raw->data=query->conds[i].rhs_val.float_val;
                    }
                }
                else query->conds[i].rhs_val.set_vec(res);
            }
        }
        check_clause(query->tables, query->conds);

        if(x->group){
            if(!x->group->conds){
                auto tmp_v=std::vector<std::shared_ptr<ast::BinaryExpr>>{};
                x->group->conds=std::make_shared<std::vector<std::shared_ptr<ast::BinaryExpr>>>(tmp_v);
            }
            get_clause(*x->group->conds,query->having_conds,true);
            check_clause(query->tables, query->having_conds);
        }
        
    } else if (auto x = std::dynamic_pointer_cast<ast::UpdateStmt>(parse)) {
        /** TODO: */
        if(!x->conds){
            auto tmp_v=std::vector<std::shared_ptr<ast::BinaryExpr>>{};
            x->conds=std::make_shared<std::vector<std::shared_ptr<ast::BinaryExpr>>>(tmp_v);
        }
        get_clause(*x->conds,query->conds,false);
        check_clause({x->tab_name}, query->conds); 
        std::vector<SetClause> v;
        for(auto cls:*x->set_clauses){
            auto tmp=*(cls.get());
            SetClause s;
            s.lhs.col_name=cls->col_name;
            s.lhs.tab_name=x->tab_name;
            s.pre_add=cls->pre_add;
            Value new_value;
            auto p=sm_manager_->db_.get_table(x->tab_name).cols.begin();
            for(;p!=sm_manager_->db_.get_table(x->tab_name).cols.end();p++){
                if(p->name==cls->col_name)break;
            }
            if(p->type==ColType::TYPE_INT){

                if(auto value=std::dynamic_pointer_cast<ast::IntLit>(cls->val)){
                    new_value.set_int(value->val);
                new_value.init_raw(sizeof(int));
                }
                else if(auto value=std::dynamic_pointer_cast<ast::FloatLit>(cls->val)){//float->int
                    new_value.set_int(int(value->val));
                new_value.init_raw(sizeof(int));
                }
            }
            else if(p->type==ColType::TYPE_FLOAT){

                if(auto value=std::dynamic_pointer_cast<ast::FloatLit>(cls->val)){
                    new_value.set_float(value->val);
                    new_value.init_raw(sizeof(float));
                }
                else if(auto value=std::dynamic_pointer_cast<ast::IntLit>(cls->val)){//int->float
                    new_value.set_float((float)value->val);
                    new_value.init_raw(sizeof(float));
                }
            }
            else if(p->type==ColType::TYPE_STRING){
                auto value=std::dynamic_pointer_cast<ast::StringLit>(cls->val);
                new_value.set_str(value->val);
                new_value.init_raw(value->val.size());
            }
            else assert(0);
            s.rhs=new_value;
            v.push_back(s);
        }
        query->set_clauses=v;
    } else if (auto x = std::dynamic_pointer_cast<ast::DeleteStmt>(parse)) {
        //处理where条件
        if(!x->conds){
            auto tmp_v=std::vector<std::shared_ptr<ast::BinaryExpr>>{};
            x->conds=std::make_shared<std::vector<std::shared_ptr<ast::BinaryExpr>>>(tmp_v);
        }
        get_clause(*x->conds, query->conds,false);
        check_clause({x->tab_name}, query->conds);        
    } else if (auto x = std::dynamic_pointer_cast<ast::InsertStmt>(parse)) {
        // 处理insert 的values值
        for (auto &sv_val : *x->vals) {
            query->values.push_back(convert_sv_value(sv_val));
        }
    } else {
        // do nothing
    }
    query->parse = std::move(parse);
    return query;
}


TabCol Analyze::check_column(const std::vector<ColMeta> &all_cols, TabCol target) {
    if (target.tab_name.empty()) {
        // Table name not specified, infer table name from column name
        std::string tab_name;
        for (auto &col : all_cols) {
            if (col.name == target.col_name) {
                if (!tab_name.empty()) {
                    throw AmbiguousColumnError(target.col_name);
                }
                tab_name = col.tab_name;
            }
        }
        if (tab_name.empty()) {
            throw ColumnNotFoundError(target.col_name);
        }
        target.tab_name = tab_name;
    } else {
        /** TODO: Make sure target column exists */
        bool legal=false;
        for(const auto&col:all_cols){
            if(col.name==target.col_name&&col.tab_name==target.tab_name){
                legal=true;
                break;
            }
        }
        if(!legal){
            throw ColumnNotFoundError(target.col_name);
        }
    }
    return target;
}

void Analyze::get_all_cols(const std::vector<std::string> &tab_names, std::vector<ColMeta> &all_cols) {
    for (auto &sel_tab_name : tab_names) {
        // 这里db_不能写成get_db(), 注意要传指针
        auto sel_tab_cols = sm_manager_->db_.get_table(sel_tab_name).cols;
        for(auto& sel_col:sel_tab_cols){
            sel_col.aggr_type=ast::AggrType::NO_AGGR;
        }
        all_cols.insert(all_cols.end(), sel_tab_cols.begin(), sel_tab_cols.end());
    }
}

void Analyze::get_clause(const std::vector<std::shared_ptr<ast::BinaryExpr>> &sv_conds, std::vector<Condition> &conds,bool in_group) {
    conds.clear();
    for (auto &expr : sv_conds) {
        Condition cond;
        cond.lhs_col.tab_name = expr->lhs->tab_name, cond.lhs_col.col_name = expr->lhs->col_name, cond.lhs_col.aggr_type=expr->lhs->aggr_type;
        if(expr->lhs->aggr_type!=ast::AggrType::NO_AGGR&&!in_group)throw IllegalGroupByError();
        cond.op = convert_sv_comp_op(expr->op);
        if(expr->v&&expr->v->size()>0){
            cond.is_rhs_vector=true;
            cond.is_rhs_val=false;
        }
        else if (auto rhs_val = std::dynamic_pointer_cast<ast::Value>(expr->rhs)) {
            cond.is_rhs_val = true;
            cond.rhs_val = convert_sv_value(rhs_val);
        } else if (auto rhs_col = std::dynamic_pointer_cast<ast::Col>(expr->rhs)) {
            if(in_group){
                throw IllegalGroupByError();
            }
            cond.is_rhs_val = false;
            cond.rhs_col = {.tab_name = rhs_col->tab_name, .col_name = rhs_col->col_name};
        }
        conds.push_back(cond);
    }
}
bool judge_incompatible_type(ColType a,ColType b){
    if(a==ColType::TYPE_INT&&b==ColType::TYPE_FLOAT)return true;
    return a==b;
}
void Analyze::check_clause(const std::vector<std::string> &tab_names, std::vector<Condition> &conds) {
    // auto all_cols = get_all_cols(tab_names);
    std::vector<ColMeta> all_cols;
    get_all_cols(tab_names, all_cols);
    // Get raw values in where clause
    for (auto &cond : conds) {
        // Infer table name from column name
        if(cond.lhs_col.col_name!="*")cond.lhs_col = check_column(all_cols, cond.lhs_col);
        if (!cond.is_rhs_val&&!cond.is_rhs_vector) {
            cond.rhs_col = check_column(all_cols, cond.rhs_col);
        }

        if(cond.lhs_col.aggr_type==ast::AggrType::COUNT){
            cond.rhs_val.init_raw(4);
            continue;
        }

        

        TabMeta &lhs_tab = sm_manager_->db_.get_table(cond.lhs_col.tab_name);
        auto lhs_col = lhs_tab.get_col(cond.lhs_col.col_name);
        ColType lhs_type = lhs_col->type;
        ColType rhs_type;

        if(cond.rhs_val.type==ColType::TYPE_VEC){
            if(cond.rhs_val.vec_->size()==0)rhs_type=lhs_type;//may need change!
            else rhs_type=cond.rhs_val.vec_->begin()->type;
            if(lhs_type==ColType::TYPE_FLOAT&&rhs_type==ColType::TYPE_INT){
                for(auto p=cond.rhs_val.vec_->begin();p!=cond.rhs_val.vec_->end();p++){
                    p->float_val=(float)(p->int_val);
                    *(float*)(p->raw->data)=p->float_val;
                }
                rhs_type=ColType::TYPE_FLOAT;
            }
            if (!judge_incompatible_type(lhs_type,rhs_type)) {
                throw IncompatibleTypeError(coltype2str(lhs_type), coltype2str(rhs_type));
            }
            continue;
        }
        if(lhs_type==ColType::TYPE_FLOAT&&cond.rhs_val.type==ColType::TYPE_INT){
            cond.rhs_val.float_val=(float)cond.rhs_val.int_val;
            cond.rhs_val.type=ColType::TYPE_FLOAT;
        }
        if (cond.is_rhs_val) {
                cond.rhs_val.init_raw(lhs_col->len);
                rhs_type = cond.rhs_val.type;
        } else {
            TabMeta &rhs_tab = sm_manager_->db_.get_table(cond.rhs_col.tab_name);
            auto rhs_col = rhs_tab.get_col(cond.rhs_col.col_name);
            rhs_type = rhs_col->type;
        }

        if (!judge_incompatible_type(lhs_type,rhs_type)&&(lhs_col->aggr_type!=ast::AggrType::COUNT||rhs_type!=ColType::TYPE_INT)) {
            throw IncompatibleTypeError(coltype2str(lhs_type), coltype2str(rhs_type));
        }
    }
}



Value Analyze::convert_sv_value(const std::shared_ptr<ast::Value> &sv_val) {
    Value val;
    if (auto int_lit = std::dynamic_pointer_cast<ast::IntLit>(sv_val)) {
        val.set_int(int_lit->val);
    } else if (auto float_lit = std::dynamic_pointer_cast<ast::FloatLit>(sv_val)) {
        val.set_float(float_lit->val);
    } else if (auto str_lit = std::dynamic_pointer_cast<ast::StringLit>(sv_val)) {
        val.set_str(str_lit->val);
    } else {
        throw InternalError("Unexpected sv value type");
    }
    return val;
}

CompOp Analyze::convert_sv_comp_op(ast::SvCompOp op) {
    std::map<ast::SvCompOp, CompOp> m = {
        {ast::SV_OP_EQ, OP_EQ}, {ast::SV_OP_NE, OP_NE}, {ast::SV_OP_LT, OP_LT},
        {ast::SV_OP_GT, OP_GT}, {ast::SV_OP_LE, OP_LE}, {ast::SV_OP_GE, OP_GE},
        {ast::SV_OP_IN, OP_IN}
    };
    return m.at(op);
}