/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#pragma once

#include "execution_defs.h"
#include "execution_manager.h"
#include "executor_abstract.h"
#include "index/ix.h"
#include "system/sm.h"

class IndexScanExecutor : public AbstractExecutor {
   private:
    std::string tab_name_;                      // 表名称
    TabMeta tab_;                               // 表的元数据
    std::vector<Condition> conds_;              // 扫描条件
    RmFileHandle *fh_;                          // 表的数据文件句柄
    std::vector<ColMeta> cols_;                 // 需要读取的字段
    size_t len_;                                // 选取出来的一条记录的长度
    std::vector<Condition> fed_conds_;          // 扫描条件，和conds_字段相同

    std::vector<std::string> index_col_names_;  // index scan涉及到的索引包含的字段
    IndexMeta index_meta_;                      // index scan涉及到的索引元数据
    std::unique_ptr<RmRecord> rec;
    Rid rid_;
    std::unique_ptr<IxScan> scan_;
    std::string index_name_;

    SmManager *sm_manager_;
    IxIndexHandle* ih;
    std::vector<Rid> rids;

   public:
    IndexScanExecutor(SmManager *sm_manager, std::string tab_name, std::vector<Condition> conds, std::vector<std::string> index_col_names,
                    Context *context) {
        sm_manager_ = sm_manager;
        context_ = context;
        tab_name_ = std::move(tab_name);
        tab_ = sm_manager_->db_.get_table(tab_name_);
        conds_ = std::move(conds);
        // index_no_ = index_no;
        index_col_names_ = index_col_names; 
        index_meta_ = *(tab_.get_index_meta(index_col_names_));
        fh_ = sm_manager_->fhs_.at(tab_name_).get();
        cols_ = tab_.cols;
        len_ = cols_.back().offset + cols_.back().len;
        index_name_=sm_manager_->get_ix_manager()->get_index_name(tab_name_,index_meta_.cols);
        ih = sm_manager_->ihs_[index_name_].get();
        
        std::map<CompOp, CompOp> swap_op = {
            {OP_EQ, OP_EQ}, {OP_NE, OP_NE}, {OP_LT, OP_GT}, {OP_GT, OP_LT}, {OP_LE, OP_GE}, {OP_GE, OP_LE},
        };

        for (auto &cond : conds_) {
            if (cond.lhs_col.tab_name != tab_name_) {
                // lhs is on other table, now rhs must be on this table
                assert(!cond.is_rhs_val && cond.rhs_col.tab_name == tab_name_);
                // swap lhs and rhs
                std::swap(cond.lhs_col, cond.rhs_col);
                cond.op = swap_op.at(cond.op);
            }
        }
        fed_conds_ = conds_;
    }

    void beginTuple() override {
        context_->lock_mgr_->lock_IS_on_table(context_->txn_,fh_->GetFd());
        Iid lower = ih->leaf_begin();
        std::vector<char*> low_value;
        std::vector<char*> high_value;
        std::map<std::string,int>mp;
        Iid upper = ih->leaf_end();
        int cnt=0;
        int tot_len=0;
        for(const auto&col:index_meta_.cols){
            tot_len+=col.len;
        }
        char* low_idx_value=new char[tot_len];
        char* high_idx_value=new char[tot_len];
        int offset=0;
        for(const auto&col:index_meta_.cols){
            mp[col.name]=cnt++;
            char* low_v=low_idx_value+offset;
            char* high_v=high_idx_value+offset;
            if(col.type==ColType::TYPE_INT){
                *(int*)high_v=2147483647;
                *(int*)low_v=-2147483647;
            }
            else if(col.type==ColType::TYPE_FLOAT){
                *(float*)high_v=std::numeric_limits<float>::max();
                *(float*)low_v=-std::numeric_limits<float>::max();;
            }
            else{
                memset(low_v,0x00,col.len);
                memset(high_v,0xFF,col.len);
            }
            low_value.push_back(low_v);
            high_value.push_back(high_v);
            offset+=col.len;
        }
        for (auto &cond : fed_conds_) {
            if(!cond.is_rhs_val||cond.op==OP_NE||mp.count(cond.lhs_col.col_name)==0){
                continue;
            }
            int idx=mp[cond.lhs_col.col_name];
            char* r_value=cond.rhs_val.raw->data;
            if(cond.op==OP_EQ){
                if(ix_compare(low_value[idx],r_value,index_meta_.cols[idx].type,index_meta_.cols[idx].len)<0){
                    memcpy(low_value[idx],r_value,index_meta_.cols[idx].len);
                };
                if(ix_compare(high_value[idx],r_value,index_meta_.cols[idx].type,index_meta_.cols[idx].len)>0){
                    memcpy(high_value[idx],r_value,index_meta_.cols[idx].len);
                };
            }
            else if(cond.op==OP_GE||cond.op==OP_GT){
                if(ix_compare(low_value[idx],r_value,index_meta_.cols[idx].type,index_meta_.cols[idx].len)<0){
                    memcpy(low_value[idx],r_value,index_meta_.cols[idx].len);
                };
            }
            else if(cond.op==OP_LE||cond.op==OP_LT){
                if(ix_compare(high_value[idx],r_value,index_meta_.cols[idx].type,index_meta_.cols[idx].len)>0){
                    memcpy(high_value[idx],r_value,index_meta_.cols[idx].len);
                };
            }
            else assert(0);
        }
        std::vector<ColType> types;
        std::vector<int>lens;
        for(const auto&col:index_meta_.cols){
            types.push_back(col.type);
            lens.push_back(col.len);
        }

        std::unique_lock<std::mutex> lock(ih->two_mode_latch_);
        while(ih->latch_mode_==2){
            ih->cv_.wait(lock);
        }
        ih->processing_num++;
        ih->latch_mode_=1;
        lock.unlock();
        if(ix_compare(low_idx_value,high_idx_value,types,lens)>0){
            low_idx_value=high_idx_value;
            lower=ih->upper_bound(low_idx_value);
        }
        else lower=ih->lower_bound(low_idx_value);
        upper=ih->upper_bound(high_idx_value);
        scan_ = std::make_unique<IxScan>(ih, lower, upper, sm_manager_->get_bpm());
        while(!scan_->is_end()){
            rids.push_back(scan_->rid());
            scan_->next();
        }
        lock.lock();
        ih->processing_num--;
        if(ih->processing_num==0){
            ih->latch_mode_=0;
            lock.unlock();
            ih->cv_.notify_all();
        }
        else lock.unlock();


        for(int i=0;i<rids.size()/2;i++){
            std::swap(rids[i],rids[rids.size()-i-1]);
        }
        Rid res_rid;
        delete[]low_idx_value;
        delete[]high_idx_value;
        // if(upper==ih->leaf_end()){
        //     res_rid.page_no=-1;//means the end of the b_plus_tree
        //     res_rid.slot_no=-1;
        // }
        // else{
        //     res_rid=ih->get_rid(upper);
        // }
        //context_->lock_mgr_->lock_shared_gap(context_->txn_,res_rid,index_name_);
       if(is_end())return;
        rid_ = rids.back();
        if(check())return;
        rids.pop_back();
        while (!is_end()) {
            rid_ = rids.back();
            if (check()) {
                break;
            }
            rids.pop_back();
        }
    }

    void nextTuple() override {
        if(is_end()){
            return;
        }
        rids.pop_back();
        while (!is_end()) {
            rid_ = rids.back();
            if (check()) {
                break;
            }
            rids.pop_back();
        }
    }

    std::unique_ptr<RmRecord> Next() override {
        rec=fh_->get_record(rid_,context_);
        return std::move(rec);
    }

    Rid &rid() override { return rid_; }

    const std::vector<ColMeta> &cols() const override {
        return cols_;
    };

    size_t tupleLen() const override{ return len_; };

    void cond_update(std::vector<Condition> conds)override{
        conds_=conds;
    }

    bool is_end()const override{return rids.empty();}

    bool check(){
        //context_->lock_mgr_->lock_shared_gap(context_->txn_,rid_,index_name_);
        rec=fh_->get_record(rid_,context_);
        for(const auto&cond:conds_){
            auto col=cond.lhs_col;
            auto col_info=get_col(cols_,col);
            auto left=rec->data+col_info->offset;
            char* right;
            if(cond.is_rhs_val){
                right=cond.rhs_val.raw->data;
            }
            else if(!cond.is_rhs_vector){
                if(cond.rhs_col.tab_name!=tab_name_){
                    continue;
                }
                else{
                    auto right_col_info=get_col(cols_,cond.rhs_col);
                    right=rec->data+right_col_info->offset;
                }
            }
            std::vector<ColType> coltype;
            std::vector<int> collen;
            coltype.push_back(col_info->type);
            collen.push_back(col_info->len);
            if(cond.op==CompOp::OP_IN){
                    bool found_one=false;
                    for(auto &r_rec:*(cond.rhs_val.vec_)){
                            if(ix_compare(left,r_rec.raw->data,coltype,collen)==0){
                                    found_one=true;
                            }
                    }
                    if(!found_one)return false;
                    continue;

            }
            auto res=ix_compare(left,right,coltype,collen);
            if (cond.op == CompOp::OP_EQ) {
                if(res)return false;
            } else if (cond.op == CompOp::OP_NE) {
                if(res==0)return false;
            } else if (cond.op == CompOp::OP_LT) {
                if(res>=0)return false;
            } else if (cond.op == CompOp::OP_GT) {
                if(res<=0)return false;
            } else if (cond.op == CompOp::OP_LE) {
                if(res>0)return false;
            } else if (cond.op == CompOp::OP_GE) {
                if(res<0)return false;
            } else {
                throw InternalError("Invalid operation");
                return false;
            }
        }
        return true;
    }
    std::string get_tab_name()override{return tab_name_;}
};