/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#pragma once
#include "execution_defs.h"
#include "execution_manager.h"
#include "executor_abstract.h"
#include "index/ix.h"
#include "system/sm.h"

class InsertExecutor : public AbstractExecutor {
   private:
    TabMeta tab_;                   // 表的元数据
    std::vector<Value> values_;     // 需要插入的数据
    RmFileHandle *fh_;              // 表的数据文件句柄
    std::string tab_name_;          // 表名称
    Rid rid_;                       // 插入的位置，由于系统默认插入时不指定位置，因此当前rid_在插入后才赋值
    SmManager *sm_manager_;

   public:
    InsertExecutor(SmManager *sm_manager, const std::string &tab_name, std::vector<Value> values, Context *context) {
        sm_manager_ = sm_manager;
        tab_ = sm_manager_->db_.get_table(tab_name);
        values_ = values;
        tab_name_ = tab_name;
        if (values.size() != tab_.cols.size()) {
            throw InvalidValueCountError();
        }
        fh_ = sm_manager_->fhs_.at(tab_name).get();
        context_ = context;
    };

    std::unique_ptr<RmRecord> Next() override {
        context_->lock_mgr_->lock_IX_on_table(context_->txn_,fh_->GetFd());
        // Make record buffer
        RmRecord rec(fh_->get_file_hdr().record_size);
        for (size_t i = 0; i < values_.size(); i++) {
            auto &col = tab_.cols[i];
            auto &val = values_[i];

            if(val.type==ColType::TYPE_INT&&col.type==ColType::TYPE_FLOAT){
                val.float_val=(float)val.int_val;
                val.type=ColType::TYPE_FLOAT;
            }

            if (col.type != val.type) {
                throw IncompatibleTypeError(coltype2str(col.type), coltype2str(val.type));
            }

            // if(tab_.cols[i].small_sign){
            //     bool unique_value=true;
            //     for(const auto &tmp_value:tab_.cols[i].values){
            //         if(check_value_eq(tmp_value,val)){
            //             unique_value=false;
            //             break;
            //         }
            //     }
            //     if(unique_value){
            //         tab_.cols[i].values.push_back(val);
            //         if(tab_.cols[i].values.size()>=100){
            //             tab_.cols[i].values.clear();
            //             tab_.cols[i].small_sign=false;
            //         }
            //     }
            // }
            val.init_raw(col.len);
            memcpy(rec.data + col.offset, val.raw->data, col.len);
        }


        // for(size_t i = 0; i < tab_.indexes.size(); ++i) {
        //     auto& index = tab_.indexes[i];
        //     auto index_name=sm_manager_->get_ix_manager()->get_index_name(tab_name_, index.cols);
        //     auto ih = sm_manager_->ihs_.at(index_name).get();
        //     char* key = new char[index.col_tot_len];
        //     int offset = 0;
        //     for(size_t i = 0; i < index.col_num; ++i) {
        //         memcpy(key + offset, rec.data + index.cols[i].offset, index.cols[i].len);
        //         offset += index.cols[i].len;
        //     }
        //     auto res=ih->lower_bound(key);
        //     Rid res_rid;
        //     if(res==ih->leaf_end()){
        //         res_rid.page_no=-1;//means the end of the b_plus_tree
        //         res_rid.slot_no=-1;
        //     }
        //     else{
        //         res_rid=ih->get_rid(res);
        //     }
        //     context_->lock_mgr_->lock_exclusive_gap(context_->txn_,res_rid,index_name);
        //     delete[]key;
        // }
        //context_->lock_mgr_->lock_IX_on_table(context_->txn_,sm_manager_->fhs_[tab_name_].get()->GetFd());

        for(size_t i = 0; i < tab_.indexes.size(); ++i) {
            auto& index = tab_.indexes[i];
            auto index_name=sm_manager_->get_ix_manager()->get_index_name(tab_name_, index.cols);
            auto ih = sm_manager_->ihs_.at(index_name).get();
            char* key = new char[index.col_tot_len];
            int offset = 0;
            for(size_t i = 0; i < index.col_num; ++i) {
                memcpy(key + offset, rec.data + index.cols[i].offset, index.cols[i].len);
                offset += index.cols[i].len;
            }
            //context_->lock_mgr_->lock_exclusive_gap(context_->txn_,rid_,index_name);
            if(ih->get_value(key,context_->txn_)){
                std::string tmp="failure\n";
                write_output(tmp,context_);
                delete []key;
                return nullptr;
            };
            delete[]key;
        }


        // Insert into record file
        rid_ = fh_->insert_record(rec.data, context_);

        auto log_record=new InsertLogRecord(context_->txn_->get_transaction_id(),rec,rid_,tab_name_);
        context_->log_mgr_->add_log_to_buffer(log_record);

        WriteRecord* writerecord=new WriteRecord(WType::INSERT_TUPLE,tab_name_,rid_);
        context_->txn_->append_write_record(writerecord);
        int illegal_idx=-1;
        // Insert into index
        for(size_t i = 0; i < tab_.indexes.size(); ++i) {
            auto& index = tab_.indexes[i];
            auto index_name=sm_manager_->get_ix_manager()->get_index_name(tab_name_, index.cols);
            auto ih = sm_manager_->ihs_.at(index_name).get();
            char* key = new char[index.col_tot_len];
            int offset = 0;
            for(size_t i = 0; i < index.col_num; ++i) {
                memcpy(key + offset, rec.data + index.cols[i].offset, index.cols[i].len);
                offset += index.cols[i].len;
            }
            //context_->lock_mgr_->lock_exclusive_gap(context_->txn_,rid_,index_name);
            if(!ih->insert_entry(key, rid_, context_->txn_)){
                assert(0);
            };
            delete[]key;
        }
        return nullptr;
    }
    Rid &rid() override { return rid_; }
};