/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#pragma once
#include "execution_defs.h"
#include "execution_manager.h"
#include "executor_abstract.h"
#include "index/ix.h"
#include "system/sm.h"

class NestedLoopJoinExecutor : public AbstractExecutor {
   private:
    std::unique_ptr<AbstractExecutor> left_;    // 左儿子节点（需要join的表）
    std::unique_ptr<AbstractExecutor> right_;   // 右儿子节点（需要join的表）
    size_t len_;                                // join后获得的每条记录的长度
    std::vector<ColMeta> cols_;                 // join后获得的记录的字段

    std::vector<Condition> fed_conds_;          // join条件
    bool isend;

   public:
    NestedLoopJoinExecutor(std::unique_ptr<AbstractExecutor> left, std::unique_ptr<AbstractExecutor> right, 
                            std::vector<Condition> conds) {
        left_ = std::move(left);
        right_ = std::move(right);
        len_ = left_->tupleLen() + right_->tupleLen();
        cols_ = left_->cols();
        auto right_cols = right_->cols();
        for (auto &col : right_cols) {
            col.offset += left_->tupleLen();
        }
        cols_.insert(cols_.end(), right_cols.begin(), right_cols.end());
        isend = false;
        fed_conds_ = std::move(conds);
    }

    void beginTuple() override {
        left_->beginTuple();
        if(left_->is_end())return;
        tot_update();
        right_->beginTuple();
        while(right_->is_end()){
            left_->nextTuple();
            if(left_->is_end())return;
            tot_update();
            right_->beginTuple();
        }
    }

    void nextTuple() override {
        if(left_->is_end())return;
        right_->nextTuple();
        while(right_->is_end()){
            left_->nextTuple();
            if(left_->is_end())return;
            tot_update();
            right_->beginTuple();
        }
    }

    std::unique_ptr<RmRecord> Next() override {
        if(left_->is_end()||right_->is_end()){
            return nullptr;
        }
        auto record = std::make_unique<RmRecord>(len_);
        memcpy(record->data,left_->Next()->data,left_->tupleLen());
        memcpy(record->data+left_->tupleLen(),right_->Next()->data,right_->tupleLen());
        return record;
    }

    Rid &rid() override { return _abstract_rid; }

    bool is_end()const override{return left_->is_end();}

    void tot_update(){
        auto new_conds=fed_conds_;
        auto left_res=left_->Next();
        for(auto col:left_->cols()){
            auto data=left_res->data+col.offset;
            Value col_value=Value();
            if(col.type==ColType::TYPE_FLOAT){
                col_value.set_float(*(float *)data);
                col_value.init_raw(col.len);
            }
            else if(col.type==ColType::TYPE_INT){
                col_value.set_int(*(int *)data);
                col_value.init_raw(col.len);
            }
            else if(col.type==ColType::TYPE_STRING){
                std::string str_data;
                for(int i=0;i<col.len;i++)str_data+=data[i];
                col_value.set_str(str_data);
                col_value.init_raw(col.len);
            }
            for(auto p=new_conds.begin();p!=new_conds.end();p++){
                if(p->lhs_col.tab_name==col.tab_name&&p->lhs_col.col_name==col.name){//this value is in left
                    if(p->is_rhs_val){//if it's const in right, then delete the condition
                        new_conds.erase(p);
                        p--;
                    }
                    else{//else, swap the condition
                        p->lhs_col=p->rhs_col;
                        p->is_rhs_val=true;
                        p->rhs_val=col_value;
                        if(p->op==OP_GT)p->op=OP_LT;
                        else if(p->op==OP_LT)p->op=OP_GT;
                        else if(p->op==OP_GE)p->op=OP_LE;
                        else if(p->op==OP_LE)p->op=OP_GE;
                    }
                }
                else if(!p->is_rhs_val&&p->rhs_col.col_name==col.tab_name&&p->rhs_col.col_name==col.name){//this value is in right
                    p->is_rhs_val=true;
                    p->rhs_val=col_value;
                }
            }
        }
        right_->cond_update(new_conds);
    }

    void cond_update(std::vector<Condition> conds)override{
        fed_conds_=conds;
        left_->cond_update(conds);
    }

    const std::vector<ColMeta> &cols() const override {
        return cols_;
    };

    size_t tupleLen() const override{ return len_; };
};