/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#pragma once

#include "execution_defs.h"
#include "execution_manager.h"
#include "executor_abstract.h"
#include "index/ix.h"
#include "system/sm.h"

class SeqScanExecutor : public AbstractExecutor {
   private:
    std::string tab_name_;              // 表的名称
    std::vector<Condition> conds_;      // scan的条件
    std::vector<Condition> origin_conds_;
    RmFileHandle *fh_;                  // 表的数据文件句柄
    std::vector<ColMeta> cols_;         // scan后生成的记录的字段
    size_t len_;                        // scan后生成的每条记录的长度
    std::vector<Condition> fed_conds_;  // 同conds_，两个字段相同
    std::unique_ptr<RmRecord> rec;
    Rid rid_;
    std::unique_ptr<RecScan> scan_;     // table_iterator
    std::vector<std::unique_ptr<RecScan>> scan_vec;
    TabMeta tab_;

    SmManager *sm_manager_;

   public:
    SeqScanExecutor(SmManager *sm_manager, std::string tab_name, std::vector<Condition> conds, Context *context) {
        sm_manager_ = sm_manager;
        tab_name_ = std::move(tab_name);
        conds_ = std::move(conds);
        origin_conds_=conds_;
        tab_ = sm_manager_->db_.get_table(tab_name_);
        
        fh_ = sm_manager_->fhs_.at(tab_name_).get();
        cols_ = tab_.cols;
        len_ = cols_.back().offset + cols_.back().len;
        context_ = context;
        fed_conds_ = conds_;
    }
    void beginTuple() override {
        context_->lock_mgr_->lock_shared_on_table(context_->txn_,fh_->GetFd());
        //context_->lock_mgr_->lock_shared_on_table(context_->txn_,sm_manager_->fhs_[tab_name_].get()->GetFd());
        // for(auto& idx:tab_.indexes){
        //     if(idx.cols[0].small_sign){
        //         for(auto value:idx.cols[0].values){

        //         }
        //     }
        // }
        scan_ = std::make_unique<RmScan>(fh_);
        while(!is_end()){
            while(scan_->is_end()&&!scan_vec.empty()){
                scan_=std::move(scan_vec.back());
                scan_vec.pop_back();
            }
            if(is_end())break;
            rid_=scan_->rid();
            if(check()){
                return;
            }
            scan_->next();
        }
    }

    void nextTuple() override {
        if(is_end())return;
        scan_->next();
        while(!is_end()){
            while(scan_->is_end()&&!scan_vec.empty()){
                scan_=std::move(scan_vec.back());
                scan_vec.pop_back();
            }
            if(is_end())break;
            rid_=scan_->rid();
            if(check()){
                return;
            }
            scan_->next();
        }
    }

    std::unique_ptr<RmRecord> Next() override {
        if(is_end()){
            return nullptr;
        }
        rec=fh_->get_record(rid_,nullptr);
        return std::move(rec);
    }

    Rid &rid() override { return rid_; }

    inline bool check(){
        rec=fh_->get_record(rid_,nullptr);
        for(const auto cond:conds_){
            auto col=cond.lhs_col;
            auto col_info=get_col(cols_,col);
            auto left=rec->data+col_info->offset;
            char* right;
            if(cond.is_rhs_val&&!cond.is_rhs_vector){
                right=cond.rhs_val.raw->data;
                // if(cond.rhs_val.type==ColType::TYPE_FLOAT&&col_info->type==ColType::TYPE_INT){
                     
                // }
            }
            else if(!cond.is_rhs_vector){
                if(cond.rhs_col.tab_name!=tab_name_){
                    continue;
                }
                else{
                    auto right_col_info=get_col(cols_,cond.rhs_col);
                    right=rec->data+right_col_info->offset;
                }
            }
            std::vector<ColType> coltype;
            std::vector<int> collen;
            coltype.push_back(col_info->type);
            collen.push_back(col_info->len);
            if(cond.op==CompOp::OP_IN){
                    bool found_one=false;
                    for(auto r_rec:*(cond.rhs_val.vec_)){
                            if(r_rec.type==ColType::TYPE_FLOAT&&col_info->type==ColType::TYPE_INT){
                                int tmp1=r_rec.float_val;
                                if(tmp1!=r_rec.float_val)continue;
                                else{
                                    *(int*)r_rec.raw->data=tmp1;
                                }
                            }
                            if(ix_compare(left,r_rec.raw->data,coltype,collen)==0){
                                    found_one=true;
                                    break;
                            }
                    }
                    if(!found_one)return false;
                    continue;

            }
            auto res=ix_compare(left,right,coltype,collen);
            if (cond.op == CompOp::OP_EQ) {
                if(res)return false;
            } else if (cond.op == CompOp::OP_NE) {
                if(res==0)return false;
            } else if (cond.op == CompOp::OP_LT) {
                if(res>=0)return false;
            } else if (cond.op == CompOp::OP_GT) {
                if(res<=0)return false;
            } else if (cond.op == CompOp::OP_LE) {
                if(res>0)return false;
            } else if (cond.op == CompOp::OP_GE) {
                if(res<0)return false;
            } else {
                throw InternalError("Invalid operation");
                return false;
            }
        }
        return true;
    }

    void cond_update(std::vector<Condition> conds)override{
        conds_=conds;
        for(auto &cond:origin_conds_){
            conds_.push_back(cond);
        }
    }

    const std::vector<ColMeta> &cols() const override {
        return cols_;
    };

    bool is_end()const override{return scan_->is_end()&&scan_vec.empty();}

    size_t tupleLen() const override{ return len_; };

    std::string get_tab_name()override{return tab_name_;}
};