/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#pragma once
#include "execution_defs.h"
#include "execution_manager.h"
#include "executor_abstract.h"
#include "index/ix.h"
#include "system/sm.h"
#include "sort_machine.h"

class SortMergeJoinExecutor : public AbstractExecutor {
   private:
    std::unique_ptr<AbstractExecutor> left_;    // 左儿子节点（需要join的表）
    std::unique_ptr<AbstractExecutor> right_;   // 右儿子节点（需要join的表）
    size_t len_;                                // join后获得的每条记录的长度
    std::vector<ColMeta> cols_;                 // join后获得的记录的字段

    std::vector<Condition> fed_conds_;          // join条件
    bool isend;
    bool use_index;
    std::unique_ptr<sort_machine> sort_mac_l;
    std::unique_ptr<sort_machine> sort_mac_r;
    std::unique_ptr<RmRecord> current_tuple;
    std::ifstream readFile_l;
    std::ifstream readFile_r;
    ColMeta l_col_,r_col_;
    int l_tuple_len,r_tuple_len;

    std::ofstream outFile_for_print;
   public:
    SortMergeJoinExecutor(std::unique_ptr<AbstractExecutor> left, std::unique_ptr<AbstractExecutor> right, 
                            std::vector<Condition> conds,bool use_index_=false) {
        left_ = std::move(left);
        right_ = std::move(right);
        fed_conds_ = std::move(conds);
        if(fed_conds_[0].lhs_col.tab_name!=left_->get_tab_name()){
            std::swap(fed_conds_[0].lhs_col,fed_conds_[0].rhs_col);
            if(fed_conds_[0].op==OP_GT)fed_conds_[0].op=OP_LT;
            else if(fed_conds_[0].op==OP_LT)fed_conds_[0].op=OP_GT;
            else if(fed_conds_[0].op==OP_GE)fed_conds_[0].op=OP_LE;
            else if(fed_conds_[0].op==OP_LE)fed_conds_[0].op=OP_GE;
        }
        outFile_for_print.open("sorted_results.txt");
        len_ = left_->tupleLen() + right_->tupleLen();
        l_tuple_len=left_->tupleLen();
        r_tuple_len=right_->tupleLen();
        cols_ = left_->cols();
        use_index=use_index_;
        auto right_cols = right_->cols();
        for (auto &col : right_cols) {
            col.offset += left_->tupleLen();
        }
        cols_.insert(cols_.end(), right_cols.begin(), right_cols.end());
        isend = false;
        
        if(left_->cols()[0].tab_name!=fed_conds_[0].lhs_col.tab_name){
            assert(0);
        }
        for(auto col:left_->cols()){
            if(col.name==fed_conds_[0].lhs_col.col_name){
                l_col_.len=col.len;
                l_col_.offset=col.offset;
                l_col_.type=col.type;
            }
        }     
        sort_mac_l=std::make_unique<sort_machine>(false,"l_sort",left_->tupleLen(),l_col_.offset,l_col_.type,l_col_.len,1000);
        if(right_->cols()[0].tab_name!=fed_conds_[0].rhs_col.tab_name){
            assert(0);
        }
        for(auto col:right_->cols()){
            if(col.name==fed_conds_[0].rhs_col.col_name){
                r_col_.len=col.len;
                r_col_.offset=col.offset;
                r_col_.type=col.type;
                break;
            }

        }
        sort_mac_r=std::make_unique<sort_machine>(false,"r_sort",right_->tupleLen(),r_col_.offset,r_col_.type,r_col_.len,5000000);
    }

    inline void record_printer(char* chr,bool is_left){
        std::vector<std::string> reses;
        if(is_left){
            std::string res;
            for(auto &col:left_->cols()){
                if(col.type==ColType::TYPE_INT){
                    res=std::to_string(*(int*)(chr+col.offset));
                }
                if(col.type==ColType::TYPE_FLOAT){
                    res=std::to_string(*(float*)(chr+col.offset));
                }
                if(col.type==TYPE_STRING){
                    res=std::string(chr+col.offset, col.len);
                    res.resize(strlen(res.c_str()));
                }
                reses.push_back(res);
            }
        }
        else{
            std::string res;
            for(auto &col:right_->cols()){
                 if(col.type==ColType::TYPE_INT){
                    res=std::to_string(*(int*)(chr+col.offset));
                }
                if(col.type==ColType::TYPE_FLOAT){
                    res=std::to_string(*(float*)(chr+col.offset));
                }
                if(col.type==TYPE_STRING){
                    res=std::string(chr+col.offset, col.len);
                    res.resize(strlen(res.c_str()));
                }
                reses.push_back(res);
            }
            
        }

        outFile_for_print<< "|";
        for (const auto &res: reses) {
            outFile_for_print<< " "<<res<<" |";
        }
        outFile_for_print<<"\n";

    }

    void beginTuple() override {
        std::vector<std::string> captions;
        for (auto col:left_->cols()) {
            if (col.new_name!="") {
                captions.push_back(col.new_name);
            } else {
                captions.push_back(col.name);
            }
        }
        outFile_for_print<<"|";
        for (const auto &caption:captions) {
            outFile_for_print<<" "<< caption<<" |";
        }
        outFile_for_print<<"\n";
        captions.clear();
        if(use_index){
            left_->beginTuple();
            auto res=left_->Next();
            while(!left_->is_end()){
                res=left_->Next();
                record_printer(res->data,true);
                left_->nextTuple();
            }

            for (auto col:right_->cols()) {
                if (col.new_name!="") {
                    captions.push_back(col.new_name);
                } else {
                    captions.push_back(col.name);
                }
            }
            outFile_for_print<<"|";
            for (const auto &caption:captions) {
                outFile_for_print<<" "<< caption<<" |";
            }
            outFile_for_print<<"\n";
            captions.clear();

            right_->beginTuple();
            while(!right_->is_end()){
                res=right_->Next();
                record_printer(res->data,false);
                right_->nextTuple();
            }
        }



        if(!use_index){
            //sleep(3);
            left_->beginTuple();
            while(!left_->is_end()){
                auto res=left_->Next();
                sort_mac_l->write_rec(res->data);
                left_->nextTuple();
            }
            sort_mac_l->sort_it();

            right_->beginTuple();
            while(!right_->is_end()){
                auto res=right_->Next();
                sort_mac_r->write_rec(res->data);
                right_->nextTuple();
            }
            sort_mac_r->sort_it();
            std::ifstream tmpFilel("l_sort.txt");
            std::ifstream tmpFiler("r_sort.txt");
            

            
            while(1){
                char* buf_for_print=new char[len_];
                tmpFilel.read(buf_for_print,l_tuple_len);
                if(tmpFilel.gcount()==0)break;
                record_printer(buf_for_print,true);
                delete[]buf_for_print;
            }

            for (auto col:right_->cols()) {
                if (col.new_name!="") {
                    captions.push_back(col.new_name);
                } else {
                    captions.push_back(col.name);
                }
            }
            outFile_for_print<<"|";
            for (const auto &caption:captions) {
                outFile_for_print<<" "<< caption<<" |";
            }
            outFile_for_print<<"\n";
            captions.clear();



            while(1){
                char* buf_for_print=new char[len_];
                tmpFiler.read(buf_for_print,r_tuple_len);
                if(tmpFiler.gcount()==0)break;
                record_printer(buf_for_print,false);
                delete[]buf_for_print;
            }
            tmpFilel.close();
            tmpFiler.close();

            std::ifstream outFilel("l_sort.txt");
            readFile_l=std::move(outFilel);
            std::ifstream outFiler("r_sort.txt");
            readFile_r=std::move(outFiler);
            nextTuple();
            
        }
        else{
            left_->beginTuple();
            right_->beginTuple();
            while(!left_->is_end()&&!right_->is_end()){
                auto l_res=left_->Next();
                auto r_res=right_->Next();

                auto com_res=ix_compare(l_res->data+l_col_.offset,r_res->data+r_col_.offset,l_col_.type,l_col_.len);
                if(com_res<0){
                    left_->nextTuple();
                }
                else if(com_res>0){
                    right_->nextTuple();
                }
                else{
                    current_tuple=std::make_unique<RmRecord>(len_);
                    memcpy(current_tuple->data,l_res->data,l_tuple_len);
                    memcpy(current_tuple->data+l_tuple_len,r_res->data,r_tuple_len);
                    return;
                }
            }
            if(!left_->is_end()){
                left_->nextTuple();
            }
            if(!right_->is_end()){
                right_->nextTuple();
            }
            isend=true;
        }
        
    }

    void nextTuple() override {
        if(use_index){
            left_->nextTuple();
            right_->nextTuple();
            if(left_->is_end()||right_->is_end()){
                outFile_for_print.close();
                isend=true;
                return;
            }
            auto l_res=left_->Next();
            auto r_res=right_->Next();
            while(1){
                auto com_res=ix_compare(l_res->data+l_col_.offset,r_res->data+r_col_.offset,l_col_.type,l_col_.len);
                if(com_res<0){
                    left_->nextTuple();
                    if(left_->is_end())break;
                    l_res=left_->Next();
                }
                else if(com_res>0){
                    right_->nextTuple();
                    if(right_->is_end())break;
                    r_res=right_->Next();
                }
                else{
                    current_tuple=std::make_unique<RmRecord>(len_);
                    memcpy(current_tuple->data,l_res->data,l_tuple_len);
                    memcpy(current_tuple->data+l_tuple_len,r_res->data,r_tuple_len);
                    return;
                }
            }
            outFile_for_print.close();
            isend=true;
        }
        else{
            auto l_tuple=std::make_unique<RmRecord>(l_tuple_len);
            readFile_l.read(l_tuple->data,l_tuple_len);
            auto r_tuple=std::make_unique<RmRecord>(r_tuple_len);
            readFile_r.read(r_tuple->data,r_tuple_len);
            auto com_res=ix_compare(l_tuple->data+l_col_.offset,r_tuple->data+r_col_.offset,l_col_.type,l_col_.len);
            while(com_res!=0){
                if(com_res<0){
                    readFile_l.read(l_tuple->data,l_tuple_len);
                    if(readFile_l.gcount()==0){
                        isend=true;
                        outFile_for_print.close();
                        return;
                    }
                }
                else{
                    readFile_r.read(r_tuple->data,r_tuple_len);
                    if(readFile_r.gcount()==0){
                        isend=true;
                        outFile_for_print.close();
                        return;
                    }
                }
                com_res=ix_compare(l_tuple->data+l_col_.offset,r_tuple->data+r_col_.offset,l_col_.type,l_col_.len);
            }
            current_tuple=std::make_unique<RmRecord>(len_);
            memcpy(current_tuple->data,l_tuple->data,l_tuple_len);
            memcpy(current_tuple->data+l_tuple_len,r_tuple->data,r_tuple_len);
        }
    }

    std::unique_ptr<RmRecord> Next() override {
        return std::move(current_tuple);
    }

    Rid &rid() override { return _abstract_rid; }

    bool is_end()const override{return isend;}

    const std::vector<ColMeta> &cols() const override {
        return cols_;
    };

    size_t tupleLen() const override{ return len_; };
};