/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */
#pragma once

#include <vector>
#include <string>
#include <memory>

enum JoinType {
    INNER_JOIN, LEFT_JOIN, RIGHT_JOIN, FULL_JOIN
};
namespace ast {

enum SvType {
    SV_TYPE_INT, SV_TYPE_FLOAT, SV_TYPE_STRING, SV_TYPE_BOOL
};

enum SvCompOp {
    SV_OP_EQ, SV_OP_NE, SV_OP_LT, SV_OP_GT, SV_OP_LE, SV_OP_GE ,SV_OP_IN
};

enum OrderByDir {
    OrderBy_DEFAULT,
    OrderBy_ASC,
    OrderBy_DESC
};

enum AggrType {
    NO_AGGR,
    MAX,
    MIN,
    COUNT,
    SUM
};


enum SetKnobType {
    EnableNestLoop, EnableSortMerge
};

// Base class for tree nodes
struct TreeNode {
    virtual ~TreeNode() = default;  // enable polymorphism
};

struct Help : public TreeNode {
};

struct ShowTables : public TreeNode {
};

struct TxnBegin : public TreeNode {
};

struct TxnCommit : public TreeNode {
};

struct TxnAbort : public TreeNode {
};

struct TxnRollback : public TreeNode {
};

struct TypeLen : public TreeNode {
    SvType type;
    int len;

    TypeLen(SvType type_, int len_) : type(type_), len(len_) {}
};

struct Field : public TreeNode {
};

struct ColDef : public Field {
    std::string col_name;
    std::shared_ptr<TypeLen> type_len;

    ColDef(std::string col_name_, std::shared_ptr<TypeLen> type_len_) :
            col_name(std::move(col_name_)), type_len(std::move(type_len_)) {}
};

struct CreateTable : public TreeNode {
    std::string tab_name;
    std::shared_ptr<std::vector<std::shared_ptr<Field>>> fields;

    CreateTable(std::string tab_name_, std::shared_ptr<std::vector<std::shared_ptr<Field>>> fields_) :
            tab_name(std::move(tab_name_)), fields(std::move(fields_)) {}
};

struct DropTable : public TreeNode {
    std::string tab_name;

    DropTable(std::string tab_name_) : tab_name(std::move(tab_name_)) {}
};

struct DescTable : public TreeNode {
    std::string tab_name;

    DescTable(std::string tab_name_) : tab_name(std::move(tab_name_)) {}
};

struct CreateIndex : public TreeNode {
    std::string tab_name;
    std::shared_ptr<std::vector<std::string>> col_names;

    CreateIndex(std::string tab_name_, std::shared_ptr<std::vector<std::string>> col_names_) :
            tab_name(std::move(tab_name_)), col_names(std::move(col_names_)) {}
};

struct ShowIndex : public TreeNode {
    std::string tab_name;

    ShowIndex(std::string tab_name_) :
            tab_name(std::move(tab_name_)) {}
};

struct DropIndex : public TreeNode {
    std::string tab_name;
    std::shared_ptr<std::vector<std::string>> col_names;

    DropIndex(std::string tab_name_, std::shared_ptr<std::vector<std::string>> col_names_) :
            tab_name(std::move(tab_name_)), col_names(std::move(col_names_)) {}
};

struct Expr : public TreeNode {
};

struct Value : public Expr {
};

struct IntLit : public Value {
    int val;

    IntLit(int val_) : val(val_) {}
};

struct FloatLit : public Value {
    float val;

    FloatLit(float val_) : val(val_) {}
};

struct StringLit : public Value {
    std::string val;

    StringLit(std::string val_) : val(std::move(val_)) {}
};

struct BoolLit : public Value {
    bool val;

    BoolLit(bool val_) : val(val_) {}
};

struct Col : public Expr {
    std::string tab_name;
    std::string col_name;
    std::string new_name;
    AggrType aggr_type=AggrType::NO_AGGR;
    Col(std::string tab_name_, std::string col_name_) :
            tab_name(std::move(tab_name_)), col_name(std::move(col_name_)) {}

    Col(std::string tab_name_, std::string col_name_, std::string new_name_):
    tab_name(std::move(tab_name_)),col_name(std::move(col_name_)),new_name(std::move(new_name_)){}

    Col(AggrType aggr_type_,std::string tab_name_, std::string col_name_, std::string new_name_):
    tab_name(std::move(tab_name_)),col_name(std::move(col_name_)),new_name(std::move(new_name_)),aggr_type(std::move(aggr_type_)){}
};

struct SetClause : public TreeNode {
    std::string col_name;
    std::shared_ptr<Value> val;
    bool pre_add=false;

    SetClause(std::string col_name_, std::shared_ptr<Value> val_) :
            col_name(std::move(col_name_)), val(std::move(val_)) {}

    SetClause(std::string col_name_, std::shared_ptr<Value> val_,bool pre_add_) :
            col_name(std::move(col_name_)), val(std::move(val_)), pre_add(pre_add_){};
};


struct BinaryExpr : public TreeNode {
    std::shared_ptr<Col> lhs;
    SvCompOp op;
    std::shared_ptr<Expr> rhs;
    std::shared_ptr<std::vector<std::shared_ptr<Value>>> v;

    BinaryExpr(std::shared_ptr<Col> lhs_, SvCompOp op_, std::shared_ptr<Expr> rhs_) :
            lhs(std::move(lhs_)), op(op_), rhs(std::move(rhs_)) {}

    BinaryExpr(std::shared_ptr<Col> lhs_, SvCompOp op_,std::shared_ptr<std::vector<std::shared_ptr<Value>>> v_) :
            lhs(std::move(lhs_)), op(op_), v(std::move(v_)) {}
};

struct OrderBy : public TreeNode
{
    std::shared_ptr<Col> cols;
    OrderByDir orderby_dir;
    OrderBy( std::shared_ptr<Col> cols_, OrderByDir orderby_dir_) :
       cols(std::move(cols_)), orderby_dir(std::move(orderby_dir_)) {}
};

struct GroupBy : public TreeNode
{
    std::shared_ptr<std::vector<std::shared_ptr<Col>>> cols;
    std::shared_ptr<std::vector<std::shared_ptr<BinaryExpr>>> conds;
    GroupBy(std::shared_ptr<std::vector<std::shared_ptr<Col>>> cols_):cols(std::move(cols_)){}
    GroupBy(std::shared_ptr<std::vector<std::shared_ptr<Col>>>cols_,
    std::shared_ptr<std::vector<std::shared_ptr<BinaryExpr>>> conds_):cols(std::move(cols_)),conds(std::move(conds_)){}
};

struct InsertStmt : public TreeNode {
    std::string tab_name;
    std::shared_ptr<std::vector<std::shared_ptr<Value>>> vals;

    InsertStmt(std::string tab_name_, std::shared_ptr<std::vector<std::shared_ptr<Value>>> vals_) :
            tab_name(std::move(tab_name_)), vals(std::move(vals_)) {}
};

struct DeleteStmt : public TreeNode {
    std::string tab_name;
    std::shared_ptr<std::vector<std::shared_ptr<BinaryExpr>>> conds;

    DeleteStmt(std::string tab_name_, std::shared_ptr<std::vector<std::shared_ptr<BinaryExpr>>> conds_) :
            tab_name(std::move(tab_name_)), conds(std::move(conds_)) {}
};

struct UpdateStmt : public TreeNode {
    std::string tab_name;
    std::shared_ptr<std::vector<std::shared_ptr<SetClause>>> set_clauses;
    std::shared_ptr<std::vector<std::shared_ptr<BinaryExpr>>> conds;

    UpdateStmt(std::string tab_name_,
               std::shared_ptr<std::vector<std::shared_ptr<SetClause>>> set_clauses_,
               std::shared_ptr<std::vector<std::shared_ptr<BinaryExpr>>> conds_) :
            tab_name(std::move(tab_name_)), set_clauses(std::move(set_clauses_)), conds(std::move(conds_)) {}
};

struct JoinExpr : public TreeNode {
    std::string left;
    std::string right;
    std::shared_ptr<std::vector<std::shared_ptr<BinaryExpr>>> conds;
    JoinType type;

    JoinExpr(std::string left_, std::string right_,
               std::shared_ptr<std::vector<std::shared_ptr<BinaryExpr>>> conds_, JoinType type_) :
            left(std::move(left_)), right(std::move(right_)), conds(std::move(conds_)), type(type_) {}
};

struct SelectStmt : public TreeNode {
    std::shared_ptr<std::vector<std::shared_ptr<Col>>> cols;
    std::shared_ptr<std::vector<std::string>> tabs;
    std::shared_ptr<std::vector<std::shared_ptr<BinaryExpr>>> conds;
    std::shared_ptr<std::vector<std::shared_ptr<JoinExpr>>> jointree;

    
    bool has_sort;
    std::shared_ptr<OrderBy> order;
    std::shared_ptr<GroupBy> group;

    SelectStmt(std::shared_ptr<std::vector<std::shared_ptr<Col>>> cols_,
               std::shared_ptr<std::vector<std::string>> tabs_,
               std::shared_ptr<std::vector<std::shared_ptr<BinaryExpr>>> conds_,
               std::shared_ptr<GroupBy> group_,
               std::shared_ptr<OrderBy> order_
               ) :
            cols(std::move(cols_)), tabs(std::move(tabs_)), conds(std::move(conds_)), 
            order(std::move(order_)),group(std::move(group_)) {
                has_sort = (bool)order;
            }
};

// set enable_nestloop
struct SetStmt : public TreeNode {
    SetKnobType set_knob_type_;
    bool bool_val_;

    SetStmt(SetKnobType &type, bool bool_value) : 
        set_knob_type_(type), bool_val_(bool_value) { }
};

// Semantic value
struct SemValue {
    int sv_int;
    float sv_float;
    std::string sv_str;
    bool sv_bool;
    OrderByDir sv_orderby_dir;
    std::shared_ptr<std::vector<std::string>> sv_strs;

    std::shared_ptr<TreeNode> sv_node;

    SvCompOp sv_comp_op;

    std::shared_ptr<TypeLen> sv_type_len;

    std::shared_ptr<Field> sv_field;
    std::shared_ptr<std::vector<std::shared_ptr<Field>>> sv_fields;

    std::shared_ptr<Expr> sv_expr;

    std::shared_ptr<Value> sv_val;
    std::shared_ptr<std::vector<std::shared_ptr<Value>>> sv_vals;

    std::shared_ptr<Col> sv_col;
    std::shared_ptr<std::vector<std::shared_ptr<Col>>> sv_cols;

    std::shared_ptr<SetClause> sv_set_clause;
    std::shared_ptr<std::vector<std::shared_ptr<SetClause>>> sv_set_clauses;

    std::shared_ptr<BinaryExpr> sv_cond;
    std::shared_ptr<std::vector<std::shared_ptr<BinaryExpr>>> sv_conds;

    std::shared_ptr<OrderBy> sv_orderby;
    std::shared_ptr<GroupBy> sv_groupby;

    SetKnobType sv_setKnobType;

    ~SemValue(){
        
    }

    // SemValue operator = (SemValue& A) {
	// 	this->sv_bool=A.sv_bool;
    //     this->sv_col=A.sv_col;
    //     this->sv_cols=std::move(A.sv_cols);
    //     this->sv_comp_op=std::move(A.sv_comp_op);
    //     this->sv_cond=std::move(A.sv_cond);
    //     this->sv_conds=std::move(A.sv_conds);
    //     this->sv_expr=std::move(A.sv_expr);
    //     this->sv_field=std::move(A.sv_field);
    //     this->sv_fields=std::move(A.sv_fields);
    //     this->sv_float=std::move(A.sv_float);
    //     this->sv_groupby=std::move(A.sv_groupby);
    //     this->sv_int=std::move(A.sv_int);
    //     this->sv_node=std::move(A.sv_node);
    //     this->sv_orderby=std::move(A.sv_orderby);
    //     this->sv_orderby_dir=std::move(A.sv_orderby_dir);
    //     this->sv_set_clause=std::move(A.sv_set_clause);
    //     this->sv_set_clauses=std::move(A.sv_set_clauses);
    //     this->sv_setKnobType=std::move(A.sv_setKnobType);
    //     this->sv_str=std::move(A.sv_str);
    //     this->sv_strs=std::move(A.sv_strs);
    //     this->sv_type_len=std::move(A.sv_type_len);
    //     this->sv_val=std::move(A.sv_val);
    //     this->sv_vals=std::move(A.sv_vals);
	// 	return *this; 
	// } 
};

extern std::shared_ptr<ast::TreeNode> parse_tree;

}

#define YYSTYPE ast::SemValue
