%{
#include "ast.h"
#include "yacc.tab.h"
#include <iostream>
#include <memory>

int yylex(YYSTYPE *yylval, YYLTYPE *yylloc);

void yyerror(YYLTYPE *locp, const char* s) {
    std::cerr << "Parser Error at line " << locp->first_line << " column " << locp->first_column << ": " << s << std::endl;
}

using namespace ast;
%}

// request a pure (reentrant) parser
%define api.pure full
// enable location in error handler
%locations
// enable verbose syntax error message
%define parse.error verbose

// keywords
%token SHOW TABLES CREATE TABLE DROP DESC INSERT INTO VALUES DELETE FROM ASC ORDER BY MAX MIN SUM COUNT GROUP HAVING IN AS
WHERE UPDATE SET SELECT INT CHAR FLOAT INDEX AND JOIN EXIT HELP TXN_BEGIN TXN_COMMIT TXN_ABORT TXN_ROLLBACK ORDER_BY ENABLE_NESTLOOP ENABLE_SORTMERGE DATETIME
// non-keywords
%token LEQ NEQ GEQ T_EOF

// type-specific tokens
%token <sv_str> IDENTIFIER VALUE_STRING
%token <sv_int> VALUE_INT
%token <sv_float> VALUE_FLOAT
%token <sv_bool> VALUE_BOOL

// specify types for non-terminal symbol
%type <sv_node> stmt dbStmt ddl dml txnStmt setStmt dml_select
%type <sv_field> field
%type <sv_fields> fieldList
%type <sv_type_len> type
%type <sv_comp_op> op
%type <sv_expr> expr
%type <sv_val> value
%type <sv_vals> valueList
%type <sv_str> tbName colName new_colName
%type <sv_strs> tableList colNameList
%type <sv_col> col haveCol
%type <sv_cols> colList selector
%type <sv_set_clause> setClause
%type <sv_set_clauses> setClauses
%type <sv_cond> condition
%type <sv_conds> whereClause optWhereClause
%type <sv_orderby>  order_clause opt_order_clause
%type <sv_orderby_dir> opt_asc_desc
%type <sv_setKnobType> set_knob_type
%type <sv_groupby> opt_group_clause

%%
start:
        stmt ';'
    {
        parse_tree = $1;
        YYACCEPT;
    }
    |   HELP
    {
        parse_tree = std::make_shared<Help>();
        YYACCEPT;
    }
    |   EXIT
    {
        parse_tree = nullptr;
        YYACCEPT;
    }
    |   T_EOF
    {
        parse_tree = nullptr;
        YYACCEPT;
    }
    ;

stmt:
        dbStmt
    |   ddl
    |   dml
    |   txnStmt
    |   setStmt
    ;

txnStmt:
        TXN_BEGIN
    {
        $$ = std::make_shared<TxnBegin>();
    }
    |   TXN_COMMIT
    {
        $$ = std::make_shared<TxnCommit>();
    }
    |   TXN_ABORT
    {
        $$ = std::make_shared<TxnAbort>();
    }
    | TXN_ROLLBACK
    {
        $$ = std::make_shared<TxnRollback>();
    }
    ;

dbStmt:
        SHOW TABLES
    {
        $$ = std::make_shared<ShowTables>();
    }
    |
        SHOW INDEX FROM tbName
    {
        $$ = std::make_shared<ShowIndex>($4);
    }

    ;

setStmt:
        SET set_knob_type '=' VALUE_BOOL
    {
        $$ = std::make_shared<SetStmt>($2, $4);
    }
    ;

ddl:
        CREATE TABLE tbName '(' fieldList ')'
    {
        $$ = std::make_shared<CreateTable>($3, $5);
    }
    |   DROP TABLE tbName
    {
        $$ = std::make_shared<DropTable>($3);
    }
    |   DESC tbName
    {
        $$ = std::make_shared<DescTable>($2);
    }
    |   CREATE INDEX tbName '(' colNameList ')'
    {
        $$ = std::make_shared<CreateIndex>($3, $5);
    }
    |   DROP INDEX tbName '(' colNameList ')'
    {
        $$ = std::make_shared<DropIndex>($3, $5);
    }
    ;

dml:
        INSERT INTO tbName VALUES '(' valueList ')'
    {
        $$ = std::make_shared<InsertStmt>($3, $6);
    }
    |   DELETE FROM tbName optWhereClause
    {
        $$ = std::make_shared<DeleteStmt>($3, $4);
    }
    |   UPDATE tbName SET setClauses optWhereClause
    {
        $$ = std::make_shared<UpdateStmt>($2, $4, $5);
    }
    |   dml_select
    {
        $$=$1;
    }
    ;

dml_select:
    SELECT selector FROM tableList optWhereClause opt_group_clause opt_order_clause
    {
        $$ = std::make_shared<SelectStmt>($2, $4, $5, $6, $7);
    }

fieldList:
        field
    {
        $$ = std::make_shared<std::vector<std::shared_ptr<Field>>>(std::vector<std::shared_ptr<Field>>{$1});
    }
    |   fieldList ',' field
    {
        $$->push_back($3);
    }
    ;

colNameList:
        colName
    {
        $$ = std::make_shared<std::vector<std::string>>(std::vector<std::string>{$1});
    }
    | colNameList ',' colName
    {
        $$->push_back($3);
    }
    ;

field:
        colName type
    {
        $$ = std::make_shared<ColDef>($1, $2);
    }
    ;

type:
        INT
    {
        $$ = std::make_shared<TypeLen>(SV_TYPE_INT, sizeof(int));
    }
    |   CHAR '(' VALUE_INT ')'
    {
        $$ = std::make_shared<TypeLen>(SV_TYPE_STRING, $3);
    }
    |   FLOAT
    {
        $$ = std::make_shared<TypeLen>(SV_TYPE_FLOAT, sizeof(float));
    }
    |   DATETIME
    {
        $$ = std::make_shared<TypeLen>(SV_TYPE_STRING,19);
    }
    ;

valueList:
        value
    {
        $$ =std::make_shared<std::vector<std::shared_ptr<Value>>>(std::vector<std::shared_ptr<Value>>{$1});
    }
    |   valueList ',' value
    {
        $$->push_back($3);
    }
    ;

value:
        VALUE_INT
    {
        $$ = std::make_shared<IntLit>($1);
    }
    |   VALUE_FLOAT
    {
        $$ = std::make_shared<FloatLit>($1);
    }
    |   VALUE_STRING
    {
        $$ = std::make_shared<StringLit>($1);
    }
    |   VALUE_BOOL
    {
        $$ = std::make_shared<BoolLit>($1);
    }
    ;

condition:
        col op expr
    {
        $$ = std::make_shared<BinaryExpr>($1, $2, $3);
    }
    | col IN '(' valueList ')'
    {
        $$=std::make_shared<BinaryExpr>($1,SV_OP_IN,$4);
    }
    ;

optWhereClause:
        /* epsilon */ { /* ignore*/ }
    |   WHERE whereClause
    {
        $$ = $2;
    }
    ;

whereClause:
        condition 
    {
        $$ =std::make_shared<std::vector<std::shared_ptr<BinaryExpr>>>(std::vector<std::shared_ptr<BinaryExpr>>{$1});
    }
    |   whereClause AND condition
    {
        $$->push_back($3);
    }
    ;


col:
        tbName '.' colName
    {
        $$ = std::make_shared<Col>($1, $3);
    }
    |   colName
    {
        $$ = std::make_shared<Col>("", $1);
    }
    |   MAX '(' colName ')' AS new_colName
    {
        $$ = std::make_shared<Col>(AggrType::MAX,"",$3,$6);
    }
    |   MAX '(' colName ')'
    {
        $$ = std::make_shared<Col>(AggrType::MAX,"",$3,"");
    }
    |   MAX '(' tbName '.' colName ')' AS new_colName
    {
        $$ = std::make_shared<Col>(AggrType::MAX,$3,$5,$8);
    }
    |   MIN '(' colName ')' AS new_colName
    {
        $$ = std::make_shared<Col>(AggrType::MIN,"",$3,$6);
    }
    |   MIN '(' tbName '.' colName ')' AS new_colName
    {
        $$ = std::make_shared<Col>(AggrType::MIN,$3,$5,$8);
    }
    |   COUNT '(' colName ')' AS new_colName
    {
        $$ = std::make_shared<Col>(AggrType::COUNT,"",$3,$6);
    }
    |   COUNT '(' tbName '.' colName ')' AS new_colName
    {
        $$ = std::make_shared<Col>(AggrType::COUNT,$3,$5,$8);
    }
    |   COUNT '(' '*' ')' AS new_colName
    {
        $$ = std::make_shared<Col>(AggrType::COUNT,"","*",$6);
    }
    |   SUM '(' colName ')' AS new_colName
    {
        $$ = std::make_shared<Col>(AggrType::SUM,"",$3,$6);
    }
    |   SUM '(' tbName '.' colName ')' AS new_colName
    {
        $$ = std::make_shared<Col>(AggrType::SUM,$3,$5,$8);
    }
    |   tbName '.' colName AS new_colName
    {
        $$ = std::make_shared<Col>($1,$3,$5);
    }
    |   colName AS new_colName
    {
        $$ = std::make_shared<Col>("",$1,$3);
    }
    |   haveCol
    {
        $$=$1;
    }
    ;

colList:
        col
    {
        $$ =std::make_shared<std::vector<std::shared_ptr<Col>>>(std::vector<std::shared_ptr<Col>>{$1});
    }
    |   colList ',' col
    {
        $$->push_back($3);
    }
    ;

haveCol:
       MAX '(' colName ')'
    {
        $$ = std::make_shared<Col>(AggrType::MAX,"",$3,"");
    }
    |   MAX '(' tbName '.' colName ')'
    {
        $$ = std::make_shared<Col>(AggrType::MAX,$3,$5,"");
    }
    |   MIN '(' colName ')'
    {
        $$ = std::make_shared<Col>(AggrType::MIN,"",$3,"");
    }
    |   MIN '(' tbName '.' colName ')'
    {
        $$ = std::make_shared<Col>(AggrType::MIN,$3,$5,"");
    }
    |   COUNT '(' colName ')'
    {
        $$ = std::make_shared<Col>(AggrType::COUNT,"",$3,"");
    }
    |   COUNT '(' tbName '.' colName ')'
    {
        $$ = std::make_shared<Col>(AggrType::COUNT,$3,$5,"");
    }
    |   COUNT '(' '*' ')'
    {
        $$ = std::make_shared<Col>(AggrType::COUNT,"","*","");
    }
    |   SUM '(' colName ')'
    {
        $$ = std::make_shared<Col>(AggrType::SUM,"",$3,"");
    }
    |   SUM '(' tbName '.' colName ')'
    {
        $$ = std::make_shared<Col>(AggrType::SUM,$3,$5,"");
    }

op:
        '='
    {
        $$ = SV_OP_EQ;
    }
    |   '<'
    {
        $$ = SV_OP_LT;
    }
    |   '>'
    {
        $$ = SV_OP_GT;
    }
    |   NEQ
    {
        $$ = SV_OP_NE;
    }
    |   LEQ
    {
        $$ = SV_OP_LE;
    }
    |   GEQ
    {
        $$ = SV_OP_GE;
    }
    ;

expr:
        value
    {
        $$ = std::static_pointer_cast<Expr>($1);
    }
    |   col
    {
        $$ = std::static_pointer_cast<Expr>($1);
    }
    ;

setClauses:
        setClause
    {
        $$ =std::make_shared<std::vector<std::shared_ptr<SetClause>>>(std::vector<std::shared_ptr<SetClause>>{$1});
    }
    |   setClauses ',' setClause
    {
        $$->push_back($3);
    }
    ;

setClause:
        colName '=' value
    {
        $$ = std::make_shared<SetClause>($1, $3);
    }
    |       colName '=' colName value
    {
        $$ = std::make_shared<SetClause>($1, $4, true);
    }
    ;

selector:
        '*'
    {
        $$ = {};
    }
    |   colList
    ;

tableList:
        tbName
    {
        $$ =std::make_shared<std::vector<std::string>>(std::vector<std::string>{$1});
    }
    |   tableList ',' tbName
    {
        $$->push_back($3);
    }
    |   tableList JOIN tbName
    {
        $$->push_back($3);
    }
    ;

opt_order_clause:
    ORDER BY order_clause      
    { 
        $$ = $3; 
    }
    |   /* epsilon */ { /* ignore*/ }
    ;

order_clause:
      col  opt_asc_desc 
    { 
        $$ = std::make_shared<OrderBy>($1, $2);
    }
    ;   

opt_group_clause:
    GROUP BY colList
    { 
        $$=std::make_shared<GroupBy>($3);
    }
    | GROUP BY colList HAVING whereClause
    {
        $$=std::make_shared<GroupBy>($3,$5);
    }
    |
    {}
    ;


opt_asc_desc:
    ASC          { $$ = OrderBy_ASC;     }
    |  DESC      { $$ = OrderBy_DESC;    }
    |       { $$ = OrderBy_DEFAULT; }
    ;    

set_knob_type:
    ENABLE_NESTLOOP { $$ = EnableNestLoop; }
    |   ENABLE_SORTMERGE { $$ = EnableSortMerge; }
    ;

tbName: IDENTIFIER;

colName: IDENTIFIER;

new_colName: IDENTIFIER;
%%
