/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#include "rm_file_handle.h"

/**
 * @description: 获取当前表中记录号为rid的记录
 * @param {Rid&} rid 记录号，指定记录的位置
 * @param {Context*} context
 * @return {unique_ptr<RmRecord>} rid对应的记录对象指针
 */
std::unique_ptr<RmRecord> RmFileHandle::get_record(const Rid& rid, Context* context) const {
    // Todo:
    // 1. 获取指定记录所在的page handle
    // 2. 初始化一个指向RmRecord的指针（赋值其内部的data和size）
    if(context)context->lock_mgr_->lock_shared_on_record(context->txn_,rid,fd_);
    auto page_handle=fetch_page_handle(rid.page_no);
    auto new_record=std::make_unique<RmRecord> (file_hdr_.record_size);
    memcpy(new_record->data,page_handle.get_slot(rid.slot_no),file_hdr_.record_size);
    new_record->size=file_hdr_.record_size;
    buffer_pool_manager_->unpin_page(page_handle.page->get_page_id(),false);
    return new_record;
}

/**
 * @description: 在当前表中插入一条记录，不指定插入位置
 * @param {char*} buf 要插入的记录的数据
 * @param {Context*} context
 * @return {Rid} 插入的记录的记录号（位置）
 */
Rid RmFileHandle::insert_record(char* buf, Context* context) {
    rm_latch_.lock();
    auto page_handle=create_page_handle();
    int free_slot=Bitmap::first_bit(0,page_handle.bitmap,file_hdr_.num_records_per_page);
    if(free_slot==file_hdr_.num_records_per_page)assert(0);
    auto rid=Rid{page_handle.page->get_page_id().page_no, free_slot};
    rm_latch_.unlock();
    if(context!=nullptr)context->lock_mgr_->lock_exclusive_on_record(context->txn_,rid,fd_);
    rm_latch_.lock();
    if(Bitmap::is_set(page_handle.bitmap,free_slot)){
        rm_latch_.unlock();
        return insert_record(buf,context);
    }
    memcpy(page_handle.get_slot(free_slot),buf,file_hdr_.record_size);
    Bitmap::set(page_handle.bitmap,free_slot);
    if(++page_handle.page_hdr->num_records==file_hdr_.num_records_per_page)
        file_hdr_.first_free_page_no=page_handle.page_hdr->next_free_page_no;
    buffer_pool_manager_->unpin_page(page_handle.page->get_page_id(),true);
    rm_latch_.unlock();
    return rid;
}

/**
 * @description: 在当前表中的指定位置插入一条记录
 * @param {Rid&} rid 要插入记录的位置
 * @param {char*} buf 要插入记录的数据
 */
void RmFileHandle::insert_record(const Rid& rid, char* buf) {
    if(rid.page_no==file_hdr_.num_pages){
        create_new_page_handle();
    }
    if (rid.page_no >= file_hdr_.num_pages) {
        assert(0);
        throw RecordNotFoundError(rid.page_no,rid.slot_no);
        return;
    }
    int page_no=rid.page_no;
    int slot_no=rid.slot_no;
    auto page_handle=fetch_page_handle(page_no);
    rm_latch_.lock();
    //if(Bitmap::is_set(page_handle.bitmap,slot_no))assert(0);
    Bitmap::set(page_handle.bitmap,slot_no);
    memcpy(page_handle.get_slot(slot_no),buf,file_hdr_.record_size);
    if(++page_handle.page_hdr->num_records==file_hdr_.num_records_per_page&&page_no==file_hdr_.first_free_page_no)file_hdr_.first_free_page_no=page_handle.page_hdr->next_free_page_no;
    buffer_pool_manager_->unpin_page(page_handle.page->get_page_id(), true);
    rm_latch_.unlock();
}

/**
 * @description: 删除记录文件中记录号为rid的记录
 * @param {Rid&} rid 要删除的记录的记录号（位置）
 * @param {Context*} context
 */
void RmFileHandle::delete_record(const Rid& rid, Context* context) {
    // Todo:
    // 1. 获取指定记录所在的page handle
    // 2. 更新page_handle.page_hdr中的数据结构
    // 注意考虑删除一条记录后页面未满的情况，需要调用release_page_handle()
    if(context)context->lock_mgr_->lock_exclusive_on_record(context->txn_,rid,fd_);
    int page_no=rid.page_no;
    int slot_no=rid.slot_no;
    auto page_handle=fetch_page_handle(page_no);
    rm_latch_.lock();
    if(slot_no>=file_hdr_.num_records_per_page||!Bitmap::is_set(page_handle.bitmap,slot_no)){
        assert(0);throw RecordNotFoundError(page_no,slot_no);
    }
    Bitmap::reset(page_handle.bitmap,slot_no);
    if(page_handle.page_hdr->num_records--==file_hdr_.num_records_per_page)release_page_handle(page_handle);
    buffer_pool_manager_->unpin_page(page_handle.page->get_page_id(),true);
    rm_latch_.unlock();
}


/**
 * @description: 更新记录文件中记录号为rid的记录
 * @param {Rid&} rid 要更新的记录的记录号（位置）
 * @param {char*} buf 新记录的数据
 * @param {Context*} context
 */
void RmFileHandle::update_record(const Rid& rid, char* buf, Context* context) {
    // Todo:
    // 1. 获取指定记录所在的page handle
    // 2. 更新记录
    if(context)context->lock_mgr_->lock_exclusive_on_record(context->txn_,rid,fd_);
    int page_no=rid.page_no;
    int slot_no=rid.slot_no;
    auto page_handle=fetch_page_handle(page_no);
    //rm_latch_.lock();
    if(!Bitmap::is_set(page_handle.bitmap,slot_no))assert(0);
    memcpy(page_handle.get_slot(slot_no),buf,file_hdr_.record_size);
    buffer_pool_manager_->unpin_page(page_handle.page->get_page_id(),true);
    //rm_latch_.unlock();
}

/**
 * 以下函数为辅助函数，仅提供参考，可以选择完成如下函数，也可以删除如下函数，在单元测试中不涉及如下函数接口的直接调用
*/
/**
 * @description: 获取指定页面的页面句柄
 * @param {int} page_no 页面号
 * @return {RmPageHandle} 指定页面的句柄
 */
RmPageHandle RmFileHandle::fetch_page_handle(int page_no) const {
    // Todo:
    // 使用缓冲池获取指定页面，并生成page_handle返回给上层
    // if page_no is invalid, throw PageNotExistError exception
    PageId pageid={fd_,page_no};
    Page* page=buffer_pool_manager_->fetch_page(pageid);//fetchpage内部已Pin
    //buffer_pool_manager_->unpin_page(pageid,false);
    return RmPageHandle(&file_hdr_,page);
}

/**
 * @description: 创建一个新的page handle
 * @return {RmPageHandle} 新的PageHandle
 */
RmPageHandle RmFileHandle::create_new_page_handle() {
    // Todo:
    // 1.使用缓冲池来创建一个新page
    // 2.更新page handle中的相关信息
    // 3.更新file_hdr_
    PageId pageid;
    pageid.fd=fd_;
    Page* page=buffer_pool_manager_->new_page(&pageid);
    if(page==nullptr)assert(0);
    auto res=RmPageHandle(&file_hdr_,page);
    //assert(file_hdr_.first_free_page_no==-1);
    res.page_hdr->next_free_page_no=file_hdr_.first_free_page_no;//page_hdr
    res.page_hdr->num_records=0;
    file_hdr_.first_free_page_no=pageid.page_no;//file_hdr_
    file_hdr_.num_pages++;
    res.page=page;//page
    Bitmap::init(res.bitmap,file_hdr_.bitmap_size);

    buffer_pool_manager_->unpin_page(page->get_page_id(),true);
    return res;
}

/**
 * @brief 创建或获取一个空闲的page handle
 *
 * @return RmPageHandle 返回生成的空闲page handle
 * @note pin the page, remember to unpin it outside!
 */
RmPageHandle RmFileHandle::create_page_handle() {
    // Todo:
    // 1. 判断file_hdr_中是否还有空闲页
    //     1.1 没有空闲页：使用缓冲池来创建一个新page；可直接调用create_new_page_handle()
    //     1.2 有空闲页：直接获取第一个空闲页
    // 2. 生成page handle并返回给上层
    if(file_hdr_.first_free_page_no!=-1){
        return fetch_page_handle(file_hdr_.first_free_page_no);
    }
    else{
        auto res=create_new_page_handle();
        return res;
    }
}

/**
 * @description: 当一个页面从没有空闲空间的状态变为有空闲空间状态时，更新文件头和页头中空闲页面相关的元数据
 */
void RmFileHandle::release_page_handle(RmPageHandle&page_handle) {
    // Todo:
    // 当page从已满变成未满，考虑如何更新：
    // 1. page_handle.page_hdr->next_free_page_no
    // 2. file_hdr_.first_free_page_no
    page_handle.page_hdr->next_free_page_no=file_hdr_.first_free_page_no;
    file_hdr_.first_free_page_no=page_handle.page->get_page_id().page_no;
    //当一个record被delete后，它的rid在事务完成前会被加写锁，
    //这个release的操作导致在这段时间内所有其他想要insert record的事务都要获取该rid并等待或回滚，因为它是第一个空闲页唯一的空闲元组。
    
}