/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#include "sm_manager.h"

#include <sys/stat.h>
#include <unistd.h>

#include <fstream>

#include "index/ix.h"
#include "record/rm.h"
#include "record_printer.h"
#include "common/common.h"

/**
 * @description: 判断是否为一个文件夹
 * @return {bool} 返回是否为一个文件夹
 * @param {string&} db_name 数据库文件名称，与文件夹同名
 */
bool SmManager::is_dir(const std::string& db_name) {
    struct stat st;
    return stat(db_name.c_str(), &st) == 0 && S_ISDIR(st.st_mode);
}

/**
 * @description: 创建数据库，所有的数据库相关文件都放在数据库同名文件夹下
 * @param {string&} db_name 数据库名称
 */
void SmManager::create_db(const std::string& db_name) {
    if (is_dir(db_name)) {
        throw DatabaseExistsError(db_name);
    }
    //为数据库创建一个子目录
    std::string cmd = "mkdir " + db_name;
    if (system(cmd.c_str()) < 0) {  // 创建一个名为db_name的目录
        throw UnixError();
    }
    if (chdir(db_name.c_str()) < 0) {  // 进入名为db_name的目录
        throw UnixError();
    }
    //创建系统目录
    DbMeta *new_db = new DbMeta();
    new_db->name_ = db_name;

    // 注意，此处ofstream会在当前目录创建(如果没有此文件先创建)和打开一个名为DB_META_NAME的文件
    std::ofstream ofs(DB_META_NAME);

    // 将new_db中的信息，按照定义好的operator<<操作符，写入到ofs打开的DB_META_NAME文件中
    ofs << *new_db;  // 注意：此处重载了操作符<<

    delete new_db;

    // 创建日志文件
    disk_manager_->create_file(LOG_FILE_NAME);

    // 回到根目录
    if (chdir("..") < 0) {
        throw UnixError();
    }
}

/**
 * @description: 删除数据库，同时需要清空相关文件以及数据库同名文件夹
 * @param {string&} db_name 数据库名称，与文件夹同名
 */
void SmManager::drop_db(const std::string& db_name) {
    if (!is_dir(db_name)) {
        throw DatabaseNotFoundError(db_name);
    }
    std::string cmd = "rm -r " + db_name;
    if (system(cmd.c_str()) < 0) {
        throw UnixError();
    }
}

/**
 * @description: 打开数据库，找到数据库对应的文件夹，并加载数据库元数据和相关文件
 * @param {string&} db_name 数据库名称，与文件夹同名
 */
void SmManager::open_db(const std::string& db_name) {
    if (!is_dir(db_name)) {
        throw DatabaseNotFoundError(db_name);
    }
    if (chdir(db_name.c_str())<0) {
        throw UnixError();
    }
    std::ifstream ifs(DB_META_NAME);
    ifs>>db_;
    for(const auto& [_,tab]:db_.tabs_){
        fhs_[tab.name]=rm_manager_->open_file(tab.name);
        auto tmp_idxes=tab.indexes;
        for(const auto& idx:tmp_idxes){
            //auto idx_name=ix_manager_->get_index_name(tab.name,idx.cols);
            //ihs_[idx_name]=ix_manager_->open_index(tab.name,idx.cols);
            //ihs_[idx_name]=ix_manager_->copy_backup_to_main_idx(tab.name,idx.cols);
            std::vector<std::string> names;
            for(auto col:idx.cols)names.push_back(col.name);
            auto idx_name=ix_manager_->get_index_name(tab.name,idx.cols);
            ihs_[idx_name]=ix_manager_->open_index(tab.name,idx.cols);
            drop_index(tab.name,names,nullptr);
            create_index(tab.name,names,nullptr);
            
        }
    }
    
}

/**
 * @description: 把数据库相关的元数据刷入磁盘中
 */
void SmManager::flush_meta() {
    // 默认清空文件
    std::ofstream ofs(DB_META_NAME);
    ofs << db_;
}

void SmManager::flush_backup_idx(){
    for(const auto&[idx_name,idx_handler]:ihs_){
        ix_manager_->flush_index(idx_handler.get());
        disk_manager_->copy_file(idx_name,idx_name+"1");
    }
}

/**
 * @description: 关闭数据库并把数据落盘
 */
void SmManager::close_db() {
    std::ofstream ofs(DB_META_NAME);
    ofs<<db_;
    db_.name_.clear();
    db_.tabs_.clear();
    for(const auto& [_,tab]:fhs_){
        rm_manager_->close_file(tab.get());
    }
    fhs_.clear();
    for(const auto&[_,tab]:ihs_){
        ix_manager_->close_index(tab.get());
    }
    ihs_.clear();
    if (chdir("..") < 0) {
        throw UnixError();
    }
}

/**
 * @description: 显示所有的表,通过测试需要将其结果写入到output.txt,详情看题目文档
 * @param {Context*} context 
 */
void SmManager::show_tables(Context* context) {
    std::string tmp="| Tables |\n";
    RecordPrinter printer(1);
    printer.print_separator(context);
    printer.print_record({"Tables"}, context);
    printer.print_separator(context);
    for (auto &entry : db_.tabs_) {
        auto &tab = entry.second;
        printer.print_record({tab.name}, context);
        tmp+="| "+tab.name+" |\n";
    }
    printer.print_separator(context);
    write_output(tmp,context);
}

/**
 * @description: 显示表的元数据
 * @param {string&} tab_name 表名称
 * @param {Context*} context 
 */
void SmManager::desc_table(const std::string& tab_name, Context* context) {
    TabMeta &tab = db_.get_table(tab_name);

    std::vector<std::string> captions = {"Field", "Type", "Index"};
    RecordPrinter printer(captions.size());
    // Print header
    printer.print_separator(context);
    printer.print_record(captions, context);
    printer.print_separator(context);
    // Print fields
    for (auto &col : tab.cols) {
        std::vector<std::string> field_info = {col.name, coltype2str(col.type),"NO"};
        printer.print_record(field_info, context);
    }
    // Print footer
    printer.print_separator(context);
}


/**
 * @description: 显示表的index
 * @param {string&} tab_name 表名称
 * @param {Context*} context 
 */
void SmManager::show_index(const std::string& tab_name, Context* context) {
    TabMeta &tab = db_.get_table(tab_name);
    std::vector<std::string> captions = {"table_name", "unique", "columns"};
    RecordPrinter printer(captions.size());
    // Print header
    // Print fields
    for (auto &idx : tab.indexes) {
        std::string idx_names="(";
        for(const auto &col:idx.cols){
            idx_names+=col.name+',';
        }
        if(idx.col_num>0){
            idx_names[idx_names.size()-1]=')';
        }
        std::vector<std::string> field_info = {tab_name,"unique",idx_names};
        printer.print_record(field_info, context);
        std::string tmp="| "+tab.name+" | unique | "+idx_names+" |\n";
        write_output(tmp,context);
    }
    // Print footer
    printer.print_separator(context);
}

/**
 * @description: 创建表
 * @param {string&} tab_name 表的名称
 * @param {vector<ColDef>&} col_defs 表的字段
 * @param {Context*} context 
 */
void SmManager::create_table(const std::string& tab_name, const std::vector<ColDef>& col_defs, Context* context) {
    if (db_.is_table(tab_name)) {
        throw TableExistsError(tab_name);
    }
    // Create table meta
    int curr_offset = 0;
    TabMeta tab;
    tab.name = tab_name;
    for (auto &col_def : col_defs) {
        ColMeta col = {.tab_name = tab_name,
                       .name = col_def.name,
                       .type = col_def.type,
                       .len = col_def.len,
                       .offset = curr_offset};
        curr_offset += col_def.len;
        tab.cols.push_back(col);
    }
    // Create & open record file
    int record_size = curr_offset;  // record_size就是col meta所占的大小（表的元数据也是以记录的形式进行存储的）
    rm_manager_->create_file(tab_name, record_size);
    db_.tabs_[tab_name] = tab;
    // fhs_[tab_name] = rm_manager_->open_file(tab_name);
    fhs_.emplace(tab_name, rm_manager_->open_file(tab_name));

    flush_meta();
}

/**
 * @description: 删除表
 * @param {string&} tab_name 表的名称
 * @param {Context*} context
 */
void SmManager::drop_table(const std::string& tab_name, Context* context) {
    auto &tab = db_.get_table(tab_name);
    rm_manager_->close_file(fhs_[tab_name].get());
    rm_manager_->destroy_file(tab_name);
    for(const auto&idx:tab.indexes){
        auto idx_name=ix_manager_->get_index_name(tab.name,idx.cols);
        ix_manager_->close_index(ihs_[idx_name].get());
        ix_manager_->destroy_index(tab_name,idx.cols);
        ihs_.erase(idx_name);
    }
    db_.tabs_.erase(tab_name);
    fhs_.erase(tab_name);

    flush_meta();
}

/**
 * @description: 创建索引
 * @param {string&} tab_name 表的名称
 * @param {vector<string>&} col_names 索引包含的字段名称
 * @param {Context*} context
 */
void SmManager::create_index(const std::string& tab_name, const std::vector<std::string>& col_names, Context* context) {
    auto &tab = db_.get_table(tab_name);
    std::vector<ColMeta> cols;
    for(const auto&col_name:col_names){
        bool found=false;
        for(const auto&col:tab.cols){
            if(col.name==col_name){
                cols.emplace_back(col);
                found=true;
                break;
            }
        }
        if(!found){
            throw ColumnNotFoundError(col_name);
            return;
        }
    }
    auto idx_name=ix_manager_->get_index_name(tab_name,cols);
    if(ihs_.count(idx_name)){
        throw IndexExistsError(tab_name,col_names);
        return;
    }
    ix_manager_->create_index(tab_name,cols);
    auto idx_handler=std::move(ix_manager_->open_index(tab_name,col_names));
    auto file_handler=fhs_[tab_name].get();
    auto it=RmScan(file_handler);
    int key_size=0;
    for(const auto& col:cols){
        key_size+=col.len;
    }
    while(!it.is_end()){
        auto rec = std::move(file_handler->get_record(it.rid(), context));
        char* key=(char *)malloc(key_size);
        int pos=0;
        for(const auto &col:cols){
            memcpy(key+pos,rec->data+col.offset,col.len);
            pos+=col.len;
        }
        idx_handler->insert_entry(key,it.rid(),context->txn_);
        it.next();
    }
    ihs_.emplace(idx_name, std::move(idx_handler));
    auto idxmeta=new IndexMeta();
    idxmeta->tab_name=tab.name;
    idxmeta->col_num=col_names.size();
    idxmeta->col_tot_len=0;
    for(const auto&col:cols){
        idxmeta->col_tot_len+=col.len;
    }
    idxmeta->cols=cols;
    tab.indexes.push_back(*idxmeta);
    delete idxmeta;

    flush_meta();
}

/**
 * @description: 删除索引
 * @param {string&} tab_name 表名称
 * @param {vector<string>&} col_names 索引包含的字段名称
 * @param {Context*} context
 */
void SmManager::drop_index(const std::string& tab_name, const std::vector<std::string>& col_names, Context* context) {
    TabMeta& tab = db_.get_table(tab_name);
    int len=col_names.size();
    int cnt=-1;
    for(auto idx:tab.indexes){
        cnt++;
        if(idx.cols.size()!=len)continue;
        for(int i=0;i<len;i++){
            if(idx.cols[i].name!=col_names[i])break;
            else if(i==len-1){
                auto idx_name=ix_manager_->get_index_name(tab_name,idx.cols);
                auto tmp=ihs_.at(idx_name).get();
                ihs_.erase(ihs_.find(idx_name));
                ix_manager_->close_index(tmp);
                ix_manager_->destroy_index(tab_name,idx.cols);
                tab.indexes.erase(tab.indexes.begin()+cnt);
                flush_meta();
                return;
            }
        }
    }
    throw IndexNotFoundError(tab_name,col_names);
}

/**
 * @description: 删除索引
 * @param {string&} tab_name 表名称
 * @param {vector<ColMeta>&} 索引包含的字段元数据
 * @param {Context*} context
 */
void SmManager::drop_index(const std::string& tab_name, const std::vector<ColMeta>& cols, Context* context) {
    auto &tab = db_.tabs_[tab_name];
    int len=cols.size();
    for(const auto&idx:tab.indexes){
        if(idx.cols.size()!=len)continue;
        for(int i=0;i<len;i++){
            if(idx.cols[i].name!=cols[i].name)break;
            else if(i==len-1){
                auto idx_name=ix_manager_->get_index_name(tab_name,idx.cols);
                ix_manager_->close_index(ihs_[idx_name].get());
                ix_manager_->destroy_index(tab_name,idx.cols);
                ihs_.erase(idx_name);
                return;
            }
        }
    }
    std::vector<std::string> names(len);
    for(auto col:cols){
        names.push_back(col.name);
    }
    throw IndexNotFoundError(tab_name,names);
}
void SmManager::rollback_insert(const std::string &tab_name, const Rid &rid, Context *context){
        auto rm_handler=fhs_[tab_name].get();
        int cnt=0;
        TabMeta& tb=db_.get_table(tab_name);
        auto record=rm_handler->get_record(rid,context);
        for(auto idx:tb.indexes){
            char* old_key=new char[idx.col_tot_len];
            int offset=0;
            for(auto col:idx.cols){
                memcpy(old_key+offset,record->data+col.offset,col.len);
                offset+=col.len;
            }
            auto idx_handler=ihs_[ix_manager_->get_index_name(tab_name,idx.cols)].get();
            bool res;
            if(context)res=idx_handler->delete_entry(old_key,context->txn_);
            else res=idx_handler->delete_entry(old_key,nullptr);
            assert(res);
            delete[] old_key;
        }
        rm_handler->delete_record(rid,context);
}
void SmManager::rollback_delete(const std::string &tab_name, const Rid &rid, const RmRecord &record, Context *context){
        auto rm_handler=fhs_[tab_name].get();
        int cnt=0;
        rm_handler->insert_record(rid,record.data);
        TabMeta& tb=db_.get_table(tab_name);
        for(auto idx:tb.indexes){
            char* new_key=new char[idx.col_tot_len];
            int offset=0;
            for(auto col:idx.cols){
                memcpy(new_key+offset,record.data+col.offset,col.len);
                offset+=col.len;
            }
            auto idx_handler=ihs_[ix_manager_->get_index_name(tab_name,idx.cols)].get();
            bool res;
            if(context)res=idx_handler->insert_entry(new_key,rid,context->txn_);
            else res=idx_handler->insert_entry(new_key,rid,nullptr);
            assert(res);
            delete[] new_key;
        }
}

void SmManager::rollback_update(const std::string &tab_name, const Rid &rid, const RmRecord &record, Context *context)
{
        auto rm_handler=fhs_[tab_name].get();

        auto pre_record=rm_handler->get_record(rid,context);

        TabMeta& tb=db_.get_table(tab_name);
        for(auto idx:tb.indexes){
            char* old_key=new char[idx.col_tot_len];
            int offset=0;
            for(auto col:idx.cols){
                memcpy(old_key+offset,pre_record->data+col.offset,col.len);
                offset+=col.len;
            }

            char* new_key=new char[idx.col_tot_len];
            offset=0;
            for(auto col:idx.cols){
                memcpy(new_key+offset,record.data+col.offset,col.len);
                offset+=col.len;
            }
            
            std::vector<ColType>col_types;
            std::vector<int>col_lens;
            for(auto col:idx.cols){
                col_types.push_back(col.type);
                col_lens.push_back(col.len);
            }

            if(ix_compare(old_key,new_key,col_types,col_lens)==0){
                delete[]old_key;
                delete[]new_key;
                continue;
            }

            auto idx_handler=ihs_[ix_manager_->get_index_name(tab_name,idx.cols)].get();
            bool res;
            if(context)res=idx_handler->delete_entry(old_key,context->txn_);
            else res=idx_handler->delete_entry(old_key,nullptr);
            assert(res);
            delete[]old_key;

            
            if(context)res=idx_handler->insert_entry(new_key,rid,context->txn_);
            else res=idx_handler->insert_entry(new_key,rid,nullptr);
            assert(res);
            delete[]new_key;
        }
        rm_handler->update_record(rid,record.data,context);
}

void SmManager::insert_for_recovery(const std::string &tab_name, const Rid &rid, const RmRecord &record){
        auto rm_handler=fhs_[tab_name].get();
        int cnt=0;
        rm_handler->insert_record(rid,record.data);
        TabMeta& tb=db_.get_table(tab_name);
        for(auto idx:tb.indexes){
            char* new_key=new char[idx.col_tot_len];
            int offset=0;
            for(auto col:idx.cols){
                memcpy(new_key+offset,record.data+col.offset,col.len);
                offset+=col.len;
            }
            auto idx_handler=ihs_[ix_manager_->get_index_name(tab_name,idx.cols)].get();
            idx_handler->insert_entry(new_key,rid,nullptr);
        }
}

void SmManager::delete_for_recovery(const std::string &tab_name, const Rid &rid, const RmRecord &record){
        auto rm_handler=fhs_[tab_name].get();
        int cnt=0;
        TabMeta& tb=db_.get_table(tab_name);
        for(auto idx:tb.indexes){
            char* old_key=new char[idx.col_tot_len];
            int offset=0;
            for(auto col:idx.cols){
                memcpy(old_key+offset,record.data+col.offset,col.len);
                offset+=col.len;
            }
            auto idx_handler=ihs_[ix_manager_->get_index_name(tab_name,idx.cols)].get();
            idx_handler->delete_entry(old_key,nullptr);
        }
        rm_handler->delete_record(rid,nullptr);
}

void SmManager::update_for_recovery(const std::string &tab_name, const Rid &rid, const RmRecord &old_record,const RmRecord &new_record)
{
        auto rm_handler=fhs_[tab_name].get();
        int cnt=0;

        TabMeta& tb=db_.get_table(tab_name);
        for(auto idx:tb.indexes){
            char* old_key=new char[idx.col_tot_len];
            int offset=0;
            for(auto col:idx.cols){
                memcpy(old_key+offset,old_record.data+col.offset,col.len);
                offset+=col.len;
            }
            auto idx_handler=ihs_[ix_manager_->get_index_name(tab_name,idx.cols)].get();
            idx_handler->delete_entry(old_key,nullptr);
        }
        
        for(auto idx:tb.indexes){
            char* new_key=new char[idx.col_tot_len];
            int offset=0;
            for(auto col:idx.cols){
                memcpy(new_key+offset,new_record.data+col.offset,col.len);
                offset+=col.len;
            }
            auto idx_handler=ihs_[ix_manager_->get_index_name(tab_name,idx.cols)].get();
            idx_handler->insert_entry(new_key,rid,nullptr);
        }
        rm_handler->update_record(rid,new_record.data,nullptr);
}

void load_by_thread(int lower,int upper,TabMeta& tab,std::string tab_name,
std::string file_name,int record_size,std::vector<RmRecord>& vec,std::mutex& mtx,
RmFileHandle* rm_handler,std::vector<IxIndexHandle*> idx_handler,int& global_cnt){
    std::vector<char*> idx_buffers;
    std::vector<char*> rec_buffers;
    RmRecord record(record_size);
    for(auto idx:tab.indexes){
            char* new_key=new char[idx.col_tot_len];
            idx_buffers.push_back(new_key);
    }
    for(auto col:tab.cols){
            char* new_col=new char[col.len];
            rec_buffers.push_back(new_col);
    }
    std::ifstream file(file_name);
    std::string line;
    std::getline(file,line);
    int tot_cnt=0;
    while (std::getline(file, line)) {
        //std::cout<<tot_cnt++<<std::endl;
        if(tot_cnt>=lower&&(tot_cnt<upper)){
            std::istringstream lineStream(line);
            std::string cell;
            int col_cnt=0;
            int col_offset=0;
            while (std::getline(lineStream, cell, ',')) {
                if(tab.cols[col_cnt].type==ColType::TYPE_INT){
                    *(int*)(rec_buffers[col_cnt])=atoi(cell.c_str());
                }
                else if(tab.cols[col_cnt].type==ColType::TYPE_FLOAT){
                    *(float*)(rec_buffers[col_cnt])=atof(cell.c_str());
                }
                else{
                    memset(rec_buffers[col_cnt], 0,tab.cols[col_cnt].len);
                    memcpy(rec_buffers[col_cnt], cell.c_str(), cell.size());
                }
                // if(tab.cols[col_cnt].small_sign){
                //         bool no_same=true;
                //         for(auto&value:tab.cols[col_cnt].values){
                //             if(ix_compare(value,rec_buffers[col_cnt],tab.cols[col_cnt].type,tab.cols[col_cnt].len)==0){
                //                 no_same=false;
                //                 break;
                //             }
                //         }
                //         if(no_same){
                //             char* tmp_store=new char[tab.cols[col_cnt].len];
                //             memcpy(tmp_store,rec_buffers[col_cnt],tab.cols[col_cnt].len);
                //             tab.cols[col_cnt].values.push_back(tmp_store);
                //             if(tab.cols[col_cnt].values.size()==100){
                //                 tab.cols[col_cnt].values.clear();
                //                 tab.cols[col_cnt].small_sign=false;
                //             }
                //         }
                //     }
                memcpy(record.data+col_offset,rec_buffers[col_cnt],tab.cols[col_cnt].len);
                col_offset+=tab.cols[col_cnt].len;
                col_cnt++;
            }
            //std::cout<<++global_cnt<<std::endl;
            mtx.lock();
            Rid rid=rm_handler->insert_record(record.data,nullptr);
            mtx.unlock();
            int cnt=0;
            for(const auto& idx:tab.indexes){
                int offset=0;
                for(const auto& col:idx.cols){
                    memcpy(idx_buffers[cnt]+offset,record.data+col.offset,col.len);
                    offset+=col.len;
                }
                idx_handler[cnt]->insert_entry_for_load(idx_buffers[cnt],rid,nullptr);
                cnt++;
            }
        }
        //tot_cnt++;
    }
    
}

void SmManager::load_data(std::string file_name,std::string tab_name){
    //std::cout<<file_name<<"!"<<tab_name<<std::endl;
    auto rm_handler=fhs_[tab_name].get();
    TabMeta& tab=db_.get_table(tab_name);
    int record_size=rm_handler->get_file_hdr().record_size;
    std::ifstream file(file_name);
    if (!file.is_open()) {
        assert(0);
    }
    std::string line;
    int tot_cnt=0;
    int num_processes = std::thread::hardware_concurrency();
    while(std::getline(file,line)){
        tot_cnt++;
    }
    file.clear(); // Clear EOF flag
    file.seekg(0); // Reset to beginning of file

    int lines_per_process = tot_cnt / num_processes;
    int remainder = tot_cnt % num_processes;
    std::vector<std::thread>threads;
    std::mutex mtx;
    std::vector<RmRecord> vec;
    int cnt=0;
    for (int i = 0; i < num_processes; ++i) {
        int lower=i*lines_per_process;
        int upper=(i+1)*lines_per_process;
        if(i==num_processes-1)upper+=remainder;
        std::vector<IxIndexHandle*> v;
        for(auto &idx:tab.indexes){
            v.push_back(ihs_[ix_manager_->get_index_name(tab_name,idx.cols)].get());
        }
        
        threads.emplace_back( std::thread(load_by_thread,lower,upper,std::ref(tab),tab_name,
        file_name,record_size,std::ref(vec),std::ref(mtx),rm_handler,v,std::ref(cnt)));
    }
    for(auto& t:threads){
        t.join();
    }

}