/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#include "lock_manager.h"

/**
 * @description: 申请行级共享锁
 * @return {bool} 加锁是否成功
 * @param {Transaction*} txn 要申请锁的事务对象指针
 * @param {Rid&} rid 加锁的目标记录ID 记录所在的表的fd
 * @param {int} tab_fd
 */
bool LockManager::lock_shared_on_record(Transaction* txn, const Rid& rid, int tab_fd) {
    if(txn->get_state()!=TransactionState::GROWING&&txn->get_state()!=TransactionState::DEFAULT){
        throw TransactionAbortException(txn->get_transaction_id(),AbortReason::LOCK_ON_SHIRINKING);
        return false;
    }
    txn->set_state(TransactionState::GROWING);
    auto lock_data_id=std::make_unique<LockDataId>(tab_fd,rid,LockDataType::RECORD);
    std::unique_lock<std::mutex> lock(latch_);
    // if(lock_table_.count(*lock_data_id)==0){
    //     lock_table_[*lock_data_id]=
    // }
    auto lock_set=txn->get_lock_set();
    if(lock_set->count(*lock_data_id)){
        //一定会是读or写锁，所以不用判断直接return
        return true;
    }
    lock_table_[*lock_data_id].waiting_cnt++;
        while(lock_table_[*lock_data_id].group_lock_mode_!=GroupLockMode::S&&
        lock_table_[*lock_data_id].group_lock_mode_!=GroupLockMode::NON_LOCK// must be writing!
    ){
        if(!lock_table_[*lock_data_id].check_legal_wait(LockMode::SHARED,txn->get_transaction_id())){
            lock_table_[*lock_data_id].waiting_cnt--;
            throw TransactionAbortException(txn->get_transaction_id(),AbortReason::DEADLOCK_PREVENTION);
            return false;
        }
        lock_table_[*lock_data_id].cv_.wait(lock);
        
    }
    lock_table_[*lock_data_id].group_lock_mode_=GroupLockMode::S;
    lock_table_[*lock_data_id].shared_lock_num_++;
    auto lock_request=LockRequest(txn->get_transaction_id(),LockMode::SHARED);
    lock_table_[*lock_data_id].request_queue_.push_back(lock_request);

    lock.unlock();
    lock_table_[*lock_data_id].cv_.notify_all();
    lock_set->insert(*lock_data_id);
    return true;
}

bool LockManager::lock_shared_gap(Transaction* txn, const Rid& rid, std::string index_name) {
    if(txn->get_state()!=TransactionState::GROWING&&txn->get_state()!=TransactionState::DEFAULT){
        throw TransactionAbortException(txn->get_transaction_id(),AbortReason::LOCK_ON_SHIRINKING);
        return false;
    }
    txn->set_state(TransactionState::GROWING);
    auto lock_data_id=std::make_unique<LockDataId>(rid,index_name,LockDataType::GAP);
    std::unique_lock<std::mutex> lock(latch_);
    auto lock_set=txn->get_lock_set();

    if(lock_set->count(*lock_data_id)){
        //一定会是读or写锁，所以不用判断直接return
        return true;
    }
    lock_table_[*lock_data_id].waiting_cnt++;
    while(lock_table_[*lock_data_id].group_lock_mode_!=GroupLockMode::S&&
        lock_table_[*lock_data_id].group_lock_mode_!=GroupLockMode::NON_LOCK// must be writing!
    ){
        if(!lock_table_[*lock_data_id].check_legal_wait(LockMode::SHARED,txn->get_transaction_id())){
            lock_table_[*lock_data_id].waiting_cnt--;
            throw TransactionAbortException(txn->get_transaction_id(),AbortReason::DEADLOCK_PREVENTION);
            return false;
        }
        lock_table_[*lock_data_id].cv_.wait(lock);
    }
    lock_table_[*lock_data_id].group_lock_mode_=GroupLockMode::S;
    lock_table_[*lock_data_id].shared_lock_num_++;
    auto lock_request=std::make_unique<LockRequest>(txn->get_transaction_id(),LockMode::SHARED);
    lock_table_[*lock_data_id].request_queue_.push_back(*lock_request);

    lock.unlock();
    lock_table_[*lock_data_id].cv_.notify_all();
    lock_set->insert(*lock_data_id);
    return true;
}

/**
 * @description: 申请行级排他锁
 * @return {bool} 加锁是否成功
 * @param {Transaction*} txn 要申请锁的事务对象指针
 * @param {Rid&} rid 加锁的目标记录ID
 * @param {int} tab_fd 记录所在的表的fd
 */
bool LockManager::lock_exclusive_on_record(Transaction* txn, const Rid& rid, int tab_fd) {
    if(txn->get_state()!=TransactionState::GROWING&&txn->get_state()!=TransactionState::DEFAULT){
        throw TransactionAbortException(txn->get_transaction_id(),AbortReason::LOCK_ON_SHIRINKING);
        return false;
    }
    txn->set_state(TransactionState::GROWING);
    auto lock_data_id=std::make_unique<LockDataId>(tab_fd,rid,LockDataType::RECORD);
    std::unique_lock<std::mutex> lock(latch_);
    auto lock_set=txn->get_lock_set();
    int my_shared_lock_cnt=0;
    if(lock_set->count(*lock_data_id)){
        for(auto i:lock_table_[*lock_data_id].request_queue_){
            //如果是写锁就return，如果是读锁则记录
            if(i.txn_id_==txn->get_transaction_id()){
                if(i.lock_mode_==LockMode::EXLUCSIVE){
                    return true;
                }
                else{
                    my_shared_lock_cnt++;//最多只有一个
                }
                break;
            }
        }
    }
    lock_table_[*lock_data_id].waiting_cnt++;
    while(
        lock_table_[*lock_data_id].group_lock_mode_!=GroupLockMode::NON_LOCK&&
        my_shared_lock_cnt<lock_table_[*lock_data_id].request_queue_.size()
    ){
        if(!lock_table_[*lock_data_id].check_legal_wait(LockMode::EXLUCSIVE,txn->get_transaction_id())){
            lock_table_[*lock_data_id].waiting_cnt--;
            throw TransactionAbortException(txn->get_transaction_id(),AbortReason::DEADLOCK_PREVENTION);
            return false;
        }
        lock_table_[*lock_data_id].cv_.wait(lock);
        
    }
    lock_table_[*lock_data_id].group_lock_mode_=GroupLockMode::X;
    if(!my_shared_lock_cnt){//no one have the lock
        auto lock_request=std::make_unique<LockRequest>(txn->get_transaction_id(),LockMode::EXLUCSIVE);
        lock_table_[*lock_data_id].request_queue_.push_back(*lock_request);
        lock_set->insert(*lock_data_id);
    }
    else{
        lock_table_[*lock_data_id].request_queue_.begin()->lock_mode_=LockMode::EXLUCSIVE;// only one S(by the same txn)
    }

    lock.unlock();
    lock_table_[*lock_data_id].cv_.notify_all();
    return true;
}


bool LockManager::lock_exclusive_gap(Transaction* txn, const Rid& rid, std::string index_name) {
    if(txn->get_state()!=TransactionState::GROWING&&txn->get_state()!=TransactionState::DEFAULT){
        throw TransactionAbortException(txn->get_transaction_id(),AbortReason::LOCK_ON_SHIRINKING);
        return false;
    }
    txn->set_state(TransactionState::GROWING);
    auto lock_data_id=std::make_unique<LockDataId>(rid,index_name,LockDataType::GAP);
    std::unique_lock<std::mutex> lock(latch_);
    auto lock_set=txn->get_lock_set();
    int my_shared_lock_cnt=0;
    if(lock_set->count(*lock_data_id)){
        for(auto i:lock_table_[*lock_data_id].request_queue_){
            //如果是写锁就return，如果是读锁则记录
            if(i.txn_id_==txn->get_transaction_id()){
                if(i.lock_mode_==LockMode::EXLUCSIVE){
                    return true;
                }
                else{
                    my_shared_lock_cnt++;//最多只有一个
                }
                break;
            }
        }
    }
    lock_table_[*lock_data_id].waiting_cnt++;
    while(
        lock_table_[*lock_data_id].group_lock_mode_!=GroupLockMode::NON_LOCK&&
        my_shared_lock_cnt<lock_table_[*lock_data_id].request_queue_.size()
    ){
        if(!lock_table_[*lock_data_id].check_legal_wait(LockMode::EXLUCSIVE,txn->get_transaction_id())){
            lock_table_[*lock_data_id].waiting_cnt--;
            throw TransactionAbortException(txn->get_transaction_id(),AbortReason::DEADLOCK_PREVENTION);
            return false;
        }
        lock_table_[*lock_data_id].cv_.wait(lock);
        
    }
    lock_table_[*lock_data_id].group_lock_mode_=GroupLockMode::X;
    if(!my_shared_lock_cnt){//no one have the lock
        auto lock_request=std::make_unique<LockRequest>(txn->get_transaction_id(),LockMode::EXLUCSIVE);
        lock_table_[*lock_data_id].request_queue_.push_back(*lock_request);
        lock_set->insert(*lock_data_id);
    }
    else{
        lock_table_[*lock_data_id].request_queue_.begin()->lock_mode_=LockMode::EXLUCSIVE;// only one S(by the same txn)
    }

    lock.unlock();
    lock_table_[*lock_data_id].cv_.notify_all();
    return true;
}


/**
 * @description: 申请表级读锁
 * @return {bool} 返回加锁是否成功
 * @param {Transaction*} txn 要申请锁的事务对象指针
 * @param {int} tab_fd 目标表的fd
 */
bool LockManager::lock_shared_on_table(Transaction* txn, int tab_fd) {
    auto lock_data_id=std::make_unique<LockDataId>(tab_fd,LockDataType::TABLE);
    if(txn->get_state()!=TransactionState::GROWING&&txn->get_state()!=TransactionState::DEFAULT){
        throw TransactionAbortException(txn->get_transaction_id(),AbortReason::LOCK_ON_SHIRINKING);
        return false;
    }
    txn->set_state(TransactionState::GROWING);
    std::unique_lock<std::mutex> lock(latch_);
    auto lock_set=txn->get_lock_set();
    int status=0;
    if(lock_set->count(*lock_data_id)){
            //如果是X或者SIX或者S则return
            //如果是IS或者IX则分别标记
            for(auto i:lock_table_[*lock_data_id].request_queue_){
                if(i.txn_id_==txn->get_transaction_id()){
                    if(i.lock_mode_==LockMode::EXLUCSIVE||i.lock_mode_==LockMode::SHARED||
                    i.lock_mode_==LockMode::S_IX){
                        return true;
                    }

                    else{
                        if(i.lock_mode_==LockMode::INTENTION_SHARED){
                            status=1;
                        }
                        else if(i.lock_mode_==LockMode::INTENTION_EXCLUSIVE){
                            status=2;
                        }
                        break;
                    }
                }
            }
    }
    lock_table_[*lock_data_id].waiting_cnt++;
    while(lock_table_[*lock_data_id].group_lock_mode_!=GroupLockMode::S&&
        lock_table_[*lock_data_id].group_lock_mode_!=GroupLockMode::IS&&
        lock_table_[*lock_data_id].group_lock_mode_!=GroupLockMode::NON_LOCK&&
        !((status==2)&&lock_table_[*lock_data_id].IX_lock_num_==1)
    ){
        if(!lock_table_[*lock_data_id].check_legal_wait(LockMode::SHARED,txn->get_transaction_id())){
            lock_table_[*lock_data_id].waiting_cnt--;
            throw TransactionAbortException(txn->get_transaction_id(),AbortReason::DEADLOCK_PREVENTION);
            return false;
        }
        lock_table_[*lock_data_id].cv_.wait(lock);
        
    }
    
    if(status==0){
        auto lock_request=std::make_unique<LockRequest>(txn->get_transaction_id(),LockMode::SHARED);
        lock_table_[*lock_data_id].request_queue_.push_back(*lock_request);
        lock_set->insert(*lock_data_id);
        lock_table_[*lock_data_id].group_lock_mode_=GroupLockMode::S;
        lock_table_[*lock_data_id].shared_lock_num_++;
    }
    else if(status==1){//used to be IS
        lock_table_[*lock_data_id].group_lock_mode_=GroupLockMode::S;
        for(auto &i:lock_table_[*lock_data_id].request_queue_){
            if(i.txn_id_==txn->get_transaction_id()){
                i.lock_mode_=LockMode::SHARED;
            }
        }
        lock_table_[*lock_data_id].shared_lock_num_++;
    }
    else{//used to be IX
        lock_table_[*lock_data_id].group_lock_mode_=GroupLockMode::SIX;
        for(auto &i:lock_table_[*lock_data_id].request_queue_){
            if(i.txn_id_==txn->get_transaction_id()){
                i.lock_mode_=LockMode::S_IX;
            }
        }
        lock_table_[*lock_data_id].shared_lock_num_++;
    }

    lock.unlock();
    lock_table_[*lock_data_id].cv_.notify_all();
    return true;
}

/**
 * @description: 申请表级写锁
 * @return {bool} 返回加锁是否成功
 * @param {Transaction*} txn 要申请锁的事务对象指针
 * @param {int} tab_fd 目标表的fd
 */
bool LockManager::lock_exclusive_on_table(Transaction* txn, int tab_fd) {
    auto lock_data_id=std::make_unique<LockDataId>(tab_fd,LockDataType::TABLE);
    if(txn->get_state()!=TransactionState::GROWING&&txn->get_state()!=TransactionState::DEFAULT){
        throw TransactionAbortException(txn->get_transaction_id(),AbortReason::LOCK_ON_SHIRINKING);
        return false;
    }
    txn->set_state(TransactionState::GROWING);
    std::unique_lock<std::mutex> lock(latch_);
    auto lock_set=txn->get_lock_set();
    int status=0;
    if(lock_set->count(*lock_data_id)){
            //如果是X则return
            //如果其他则同一标记（无区别）
            for(auto i:lock_table_[*lock_data_id].request_queue_){
                if(i.txn_id_==txn->get_transaction_id()){
                    if(i.lock_mode_==LockMode::EXLUCSIVE){
                        return true;
                    }
                    else{
                        status=1;
                        break;
                    }
                }
            }
    }
    lock_table_[*lock_data_id].waiting_cnt++;
    while(lock_table_[*lock_data_id].group_lock_mode_!=GroupLockMode::NON_LOCK&&
        status<lock_table_[*lock_data_id].request_queue_.size()
    ){
        if(!lock_table_[*lock_data_id].check_legal_wait(LockMode::EXLUCSIVE,txn->get_transaction_id())){
            lock_table_[*lock_data_id].waiting_cnt--;
            throw TransactionAbortException(txn->get_transaction_id(),AbortReason::DEADLOCK_PREVENTION);
            return false;
        }
        lock_table_[*lock_data_id].cv_.wait(lock);
        
    }
    
    if(status==0){
        auto lock_request=LockRequest(txn->get_transaction_id(),LockMode::EXLUCSIVE);
        lock_table_[*lock_data_id].request_queue_.push_back(lock_request);
        lock_set->insert(*lock_data_id);
        lock_table_[*lock_data_id].group_lock_mode_=GroupLockMode::X;
    }
    else if(status==1){
        lock_table_[*lock_data_id].group_lock_mode_=GroupLockMode::X;
        for(auto &i:lock_table_[*lock_data_id].request_queue_){
            if(i.txn_id_==txn->get_transaction_id()){
                if(i.lock_mode_==LockMode::INTENTION_EXCLUSIVE){
                    lock_table_[*lock_data_id].IX_lock_num_--;
                }
                else if(i.lock_mode_==LockMode::SHARED){
                    lock_table_[*lock_data_id].shared_lock_num_--;
                }
                else if(i.lock_mode_==LockMode::S_IX){
                    lock_table_[*lock_data_id].IX_lock_num_--;
                    lock_table_[*lock_data_id].shared_lock_num_--;
                }
                i.lock_mode_=LockMode::EXLUCSIVE;
            }
        }
    }

    lock.unlock();
    lock_table_[*lock_data_id].cv_.notify_all();
    return true;
}

/**
 * @description: 申请表级意向读锁
 * @return {bool} 返回加锁是否成功
 * @param {Transaction*} txn 要申请锁的事务对象指针
 * @param {int} tab_fd 目标表的fd
 */
bool LockManager::lock_IS_on_table(Transaction* txn, int tab_fd) {
    auto lock_data_id=std::make_unique<LockDataId>(tab_fd,LockDataType::TABLE);
    if(txn->get_state()!=TransactionState::GROWING&&txn->get_state()!=TransactionState::DEFAULT){
        throw TransactionAbortException(txn->get_transaction_id(),AbortReason::LOCK_ON_SHIRINKING);
        return false;
    }
    txn->set_state(TransactionState::GROWING);
    std::unique_lock<std::mutex> lock(latch_);
    auto lock_set=txn->get_lock_set();
    if(lock_set->count(*lock_data_id)){
            //任意级别都>=IS,可以直接return
            return true;
    }
    lock_table_[*lock_data_id].waiting_cnt++;
    while(lock_table_[*lock_data_id].group_lock_mode_==GroupLockMode::X
    ){
        if(!lock_table_[*lock_data_id].check_legal_wait(LockMode::INTENTION_SHARED,txn->get_transaction_id())){
            lock_table_[*lock_data_id].waiting_cnt--;
            throw TransactionAbortException(txn->get_transaction_id(),AbortReason::DEADLOCK_PREVENTION);
            return false;
        }
        lock_table_[*lock_data_id].cv_.wait(lock);
        
    }
    auto lock_request=LockRequest(txn->get_transaction_id(),LockMode::INTENTION_SHARED);
    lock_table_[*lock_data_id].request_queue_.push_back(lock_request);
    lock_set->insert(*lock_data_id);
    if(lock_table_[*lock_data_id].group_lock_mode_==GroupLockMode::NON_LOCK)
        lock_table_[*lock_data_id].group_lock_mode_=GroupLockMode::IS;

    lock.unlock();
    lock_table_[*lock_data_id].cv_.notify_all();
    return true;
}

/**
 * @description: 申请表级意向写锁
 * @return {bool} 返回加锁是否成功
 * @param {Transaction*} txn 要申请锁的事务对象指针
 * @param {int} tab_fd 目标表的fd
 */
bool LockManager::lock_IX_on_table(Transaction* txn, int tab_fd) {
    auto lock_data_id=std::make_unique<LockDataId>(tab_fd,LockDataType::TABLE);
    if(txn->get_state()!=TransactionState::GROWING&&txn->get_state()!=TransactionState::DEFAULT){
        throw TransactionAbortException(txn->get_transaction_id(),AbortReason::LOCK_ON_SHIRINKING);
        return false;
    }
    txn->set_state(TransactionState::GROWING);
    std::unique_lock<std::mutex> lock(latch_);
    auto lock_set=txn->get_lock_set();
    int status=0;
    if(lock_set->count(*lock_data_id)){
            //如果是IX或者SIX或者X则return
            //如果是IS或者S则分别标记
            for(auto i:lock_table_[*lock_data_id].request_queue_){
                if(i.txn_id_==txn->get_transaction_id()){
                    if(i.lock_mode_==LockMode::EXLUCSIVE||i.lock_mode_==LockMode::INTENTION_EXCLUSIVE||
                    i.lock_mode_==LockMode::S_IX){
                        return true;
                    }
                    else{
                        if(i.lock_mode_==LockMode::INTENTION_SHARED){
                            status=1;
                        }
                        else if(i.lock_mode_==LockMode::SHARED){
                            status=2;
                        }
                        break;
                    }
                }
            }
    }
    lock_table_[*lock_data_id].waiting_cnt++;

    while(lock_table_[*lock_data_id].group_lock_mode_!=GroupLockMode::IX&&
        lock_table_[*lock_data_id].group_lock_mode_!=GroupLockMode::IS&&
        lock_table_[*lock_data_id].group_lock_mode_!=GroupLockMode::NON_LOCK&&
        !(status==2&&lock_table_[*lock_data_id].shared_lock_num_==1)
    ){
        if(!lock_table_[*lock_data_id].check_legal_wait(LockMode::INTENTION_EXCLUSIVE,txn->get_transaction_id())){
            lock_table_[*lock_data_id].waiting_cnt--;
            throw TransactionAbortException(txn->get_transaction_id(),AbortReason::DEADLOCK_PREVENTION);
            return false;
        }
        lock_table_[*lock_data_id].cv_.wait(lock);
        
    }
    
    if(status==0){
        auto lock_request=LockRequest(txn->get_transaction_id(),LockMode::INTENTION_EXCLUSIVE);
        lock_table_[*lock_data_id].request_queue_.push_back(lock_request);
        lock_set->insert(*lock_data_id);
        lock_table_[*lock_data_id].group_lock_mode_=GroupLockMode::IX;
        lock_table_[*lock_data_id].IX_lock_num_++;
    }
    else if(status==1){
        lock_table_[*lock_data_id].group_lock_mode_=GroupLockMode::IX;

        for(auto &i:lock_table_[*lock_data_id].request_queue_){
            if(i.txn_id_==txn->get_transaction_id()){
                i.lock_mode_=LockMode::INTENTION_EXCLUSIVE;
            }
        }
        lock_table_[*lock_data_id].IX_lock_num_++;
    }
    else{
        lock_table_[*lock_data_id].group_lock_mode_=GroupLockMode::SIX;
        for(auto &i:lock_table_[*lock_data_id].request_queue_){
            if(i.txn_id_==txn->get_transaction_id()){
                i.lock_mode_=LockMode::S_IX;
            }
        }
        lock_table_[*lock_data_id].IX_lock_num_++;
    }
    lock.unlock();
    lock_table_[*lock_data_id].cv_.notify_all();
    
    return true;
}

/**
 * @description: 释放锁
 * @return {bool} 返回解锁是否成功
 * @param {Transaction*} txn 要释放锁的事务对象指针
 * @param {LockDataId} lock_data_id 要释放的锁ID
 */
bool LockManager::unlock(Transaction* txn, LockDataId lock_data_id) {
   if(txn->get_state()!=TransactionState::SHRINKING&&txn->get_state()!=TransactionState::GROWING){
        throw TransactionAbortException(txn->get_transaction_id(),AbortReason::LOCK_ON_SHIRINKING);
        return false;
    }
    txn->set_state(TransactionState::SHRINKING);
    std::unique_lock<std::mutex> lock(latch_);
    
    auto lock_set=txn->get_lock_set();
    if(!lock_set->count(lock_data_id)){
        assert(0);
    }
    std::vector<LockManager::LockRequest> v;
    for(auto i:lock_table_[lock_data_id].request_queue_){
        if(i.txn_id_==txn->get_transaction_id()){
            if(i.lock_mode_==LockMode::EXLUCSIVE){
                lock_table_[lock_data_id].group_lock_mode_=GroupLockMode::NON_LOCK;
            }
            else if(i.lock_mode_==LockMode::INTENTION_EXCLUSIVE){
                lock_table_[lock_data_id].IX_lock_num_--;
            }
            else if(i.lock_mode_==LockMode::SHARED){
                lock_table_[lock_data_id].shared_lock_num_--;
            }
            else if(i.lock_mode_==LockMode::INTENTION_SHARED){
            }
            else if(i.lock_mode_==LockMode::S_IX){
                lock_table_[lock_data_id].IX_lock_num_--;
                lock_table_[lock_data_id].shared_lock_num_--;
            }
        }
        else{
            v.push_back(i);
        }
    }
    lock_table_[lock_data_id].request_queue_.clear();
    for(auto i:v){
        lock_table_[lock_data_id].request_queue_.push_back(i);
    }
    lock_table_[lock_data_id].waiting_cnt--;
    lock_table_[lock_data_id].group_lock_mode_=GroupLockMode::NON_LOCK;
    for(auto& rq:lock_table_[lock_data_id].request_queue_){
        if(rq.lock_mode_==LockMode::INTENTION_SHARED){
            if(lock_table_[lock_data_id].group_lock_mode_==GroupLockMode::NON_LOCK){
                lock_table_[lock_data_id].group_lock_mode_=GroupLockMode::IS;
            }
        }
        else if(rq.lock_mode_==LockMode::SHARED){
            if(lock_table_[lock_data_id].group_lock_mode_==GroupLockMode::NON_LOCK||
            lock_table_[lock_data_id].group_lock_mode_==GroupLockMode::IS){
                lock_table_[lock_data_id].group_lock_mode_=GroupLockMode::S;
            }
        }
        else if(rq.lock_mode_==LockMode::INTENTION_EXCLUSIVE){
            if(lock_table_[lock_data_id].group_lock_mode_==GroupLockMode::NON_LOCK||
            lock_table_[lock_data_id].group_lock_mode_==GroupLockMode::IS){
                lock_table_[lock_data_id].group_lock_mode_=GroupLockMode::IX;
            }
        }
        else if(rq.lock_mode_==LockMode::S_IX){
            if(lock_table_[lock_data_id].group_lock_mode_!=GroupLockMode::X){
                lock_table_[lock_data_id].group_lock_mode_=GroupLockMode::SIX;
            }
        }
        else if(rq.lock_mode_==LockMode::EXLUCSIVE){
            lock_table_[lock_data_id].group_lock_mode_=GroupLockMode::X;
        }
    }
    // if(lock_table_[lock_data_id].waiting_cnt==0){
    //     lock_table_.erase(lock_data_id);
    //     lock.unlock();
    //     return true;
    // }
    lock.unlock();
    lock_table_[lock_data_id].cv_.notify_all();
    return true;
}